"""This module contains SPM utilities."""
import warnings
from enum import Enum
from os import PathLike
from pathlib import Path

__all__ = [
    "SPMTissue",
    "get_spm_tissue_from_index",
    "get_tpm",
    "use_spm_standalone_if_available",
    "configure_nipype_interface_to_work_with_spm",
    "configure_nipype_interface_to_work_with_spm_standalone",
]


class SPMTissue(str, Enum):
    GRAY_MATTER = "graymatter"
    WHITE_MATTER = "whitematter"
    CSF = "csf"
    BONE = "bone"
    SOFT_TISSUE = "softtissue"
    BACKGROUND = "background"


def get_spm_tissue_from_index(index: int) -> SPMTissue:
    if index == 1:
        return SPMTissue.GRAY_MATTER
    if index == 2:
        return SPMTissue.WHITE_MATTER
    if index == 3:
        return SPMTissue.CSF
    if index == 4:
        return SPMTissue.BONE
    if index == 5:
        return SPMTissue.SOFT_TISSUE
    if index == 6:
        return SPMTissue.BACKGROUND
    raise ValueError(f"No SPM tissue matching index {index}.")


def get_tpm() -> PathLike:
    """Get Tissue Probability Map (TPM) from SPM.
    Returns
    -------
    PathLike :
        TPM.nii path from SPM
    """
    from glob import glob

    from .check_dependency import get_spm_home

    spm_home = get_spm_home()
    tpm_file_glob = glob(str(spm_home / "**/TPM.nii"), recursive=True)
    if len(tpm_file_glob) == 0:
        raise RuntimeError(f"No file found for TPM.nii in your $SPM_HOME in {spm_home}")
    if len(tpm_file_glob) > 1:
        error_str = f"Multiple files found for TPM.nii in your SPM_HOME {spm_home}:"
        for file in tpm_file_glob:
            error_str += "\n\t" + file
        raise RuntimeError(error_str)

    return tpm_file_glob[0]


def use_spm_standalone_if_available() -> bool:
    """Use SPM Standalone with MATLAB Common Runtime if it can be used on the user system.

    If there is something wrong with either SPM Standalone or the MCR environment variables
    configuration, a warning is given to the user, and False is returned.
    It is thus possible to use this function to try using the standalone version of SPM12
    while potentially defaulting to the standard SPM12 install.

    Returns
    -------
    bool :
        True if spm standalone was found and successfully configured.
        False otherwise.

    Raises
    ------
    ClinicaEnvironmentVariableError :
        If the environment variables are set to non-existent folders.
    """
    from .exceptions import ClinicaMissingDependencyError
    from .stream import log_and_warn

    try:
        configure_nipype_interface_to_work_with_spm_standalone()
        return True
    except ClinicaMissingDependencyError:
        log_and_warn(
            (
                "SPM standalone is not available on this system. "
                "The pipeline will try to use SPM and Matlab instead. "
                "If you want to rely on spm standalone, please make sure "
                "to set the following environment variables: "
                "$SPMSTANDALONE_HOME, and $MCR_HOME"
            ),
            UserWarning,
        )
        configure_nipype_interface_to_work_with_spm()
        return False


def configure_nipype_interface_to_work_with_spm() -> None:
    import nipype.interfaces.matlab as mlab

    from clinica.utils.stream import cprint

    from .check_dependency import get_spm_home

    cprint(f"Setting SPM path to {get_spm_home()}", lvl="info")
    mlab.MatlabCommand.set_default_paths(f"{get_spm_home()}")


def _get_platform_dependant_matlab_command_for_spm_standalone(
    spm_standalone_home: Path, mcr_home: Path
) -> str:
    import platform

    user_system = platform.system().lower()
    if user_system.startswith("darwin"):
        return f"cd {spm_standalone_home} && ./run_spm12.sh {mcr_home} script"
    if user_system.startswith("linux"):
        return f"{spm_standalone_home / 'run_spm12.sh'} {mcr_home} script"
    raise SystemError(
        f"Clinica only support macOS and Linux. Your system is {user_system}."
    )


def configure_nipype_interface_to_work_with_spm_standalone() -> None:
    from nipype.interfaces import spm

    from clinica.utils.stream import cprint

    from .check_dependency import get_mcr_home, get_spm_standalone_home

    spm_standalone_home = get_spm_standalone_home()
    mcr_home = get_mcr_home()
    cprint(
        f"SPM standalone has been found at {spm_standalone_home}, "
        f"with an MCR at {mcr_home} and will be used in this pipeline"
    )
    spm.SPMCommand.set_mlab_paths(
        matlab_cmd=_get_platform_dependant_matlab_command_for_spm_standalone(
            spm_standalone_home, mcr_home
        ),
        use_mcr=True,
    )
    cprint(f"Using SPM standalone version {spm.SPMCommand().version}", lvl="info")



from pathlib import Path
from typing import List

from nipype import config

from clinica.pipelines.engine import Pipeline

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class T1FreeSurfer(Pipeline):
    """FreeSurfer-based processing of T1-weighted MR images.

    Returns:
        A clinica pipeline object containing the T1FreeSurfer pipeline.
    """

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader

        image_ids: List[str] = []
        if caps_directory.is_dir():
            t1_freesurfer_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, T1_FS_DESTRIEUX, False
            )
            image_ids = extract_image_ids(t1_freesurfer_files)
        return image_ids

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.stream import cprint

        if "-dontrun" in self.parameters["recon_all_args"].split(" "):
            cprint(
                msg=(
                    "Found -dontrun flag for FreeSurfer recon-all. "
                    "Please note that this will not run the segmentation."
                ),
                lvl="warning",
            )

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Notes
        -----
        The list of inputs of the T1FreeSurfer pipeline is:
            * t1w (str): Path of the T1 weighted image in BIDS format

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["t1w"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Notes
        -----
        The list of outputs of the T1FreeSurfer pipeline is:
            * image_id (str): Image ID (e.g. sub-CLNC01_ses-M000)

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return ["image_id"]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline.

        Raise:
            ClinicaBIDSError: If there are duplicated files or missing files for any subject
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.iotools.utils.data_handling import (
            check_volume_location_in_world_coordinate_system,
        )
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.filemanip import (
            extract_subjects_sessions_from_filename,
            save_participants_sessions,
        )
        from clinica.utils.input_files import T1W_NII
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        processed_ids = self.get_processed_images(
            self.caps_directory, self.subjects, self.sessions
        )
        if len(processed_ids) > 0:
            cprint(
                msg=(
                    f"Clinica found {len(processed_ids)} image(s) "
                    "already processed in CAPS directory:"
                ),
                lvl="warning",
            )
            for image_id in processed_ids:
                cprint(msg=f"{image_id.replace('_', ' | ')}", lvl="warning")
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<session_id>/t1/freesurfer_cross_sectional"
                cprint(
                    msg=f"Output folders in {output_folder} will be recreated.",
                    lvl="warning",
                )
            else:
                cprint(msg="Image(s) will be ignored by Clinica.", lvl="warning")
                input_ids = [
                    f"{p_id}_{s_id}" for p_id, s_id in zip(self.subjects, self.sessions)
                ]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions = extract_subjects_sessions_from_filename(
                    to_process_ids
                )

        t1w_files, error_message = clinica_file_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            T1W_NII,
            raise_exception=False,
        )

        if error_message:
            cprint(error_message, lvl="warning")

        if not t1w_files:
            raise ClinicaException("Empty dataset or already processed")

        save_participants_sessions(
            self.subjects, self.sessions, self.base_dir / self.name
        )
        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {self.base_dir / self.name / 'participants.tsv'}"
            )
            cprint("The pipeline will last approximately 10 hours per image.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("t1w", t1w_files),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        check_volume_location_in_world_coordinate_system(
            t1w_files,
            self.bids_directory,
            skip_question=self.parameters["skip_question"],
        )

        self.connect(
            [
                (read_node, self.input_node, [("t1w", "t1w")]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .tasks import save_to_caps_task

        save_to_caps = npe.Node(
            interface=nutil.Function(
                input_names=["source_dir", "image_id", "caps_dir", "overwrite_caps"],
                output_names=["image_id"],
                function=save_to_caps_task,
            ),
            name="SaveToCaps",
        )
        save_to_caps.inputs.source_dir = str(self.base_dir / self.name / "ReconAll")
        save_to_caps.inputs.caps_dir = str(self.caps_directory)
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps

        self.connect(
            [
                (self.output_node, save_to_caps, [("image_id", "image_id")]),
            ]
        )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.freesurfer.preprocess import ReconAll

        from .tasks import init_input_node_task, write_tsv_files_task

        # Nodes declaration
        # =================
        # Initialize the pipeline
        #   - Extract <image_id> (e.g. sub-CLNC01_ses-M000) T1w filename;
        #   - Check FOV of T1w;
        #   - Create <subjects_dir> folder in <WD>/<Pipeline.name>/ReconAll/<image_id>/;
        #   - Print begin execution message.
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=["t1w", "recon_all_args", "output_dir"],
                output_names=["image_id", "t1w", "flags", "subjects_dir"],
                function=init_input_node_task,
            ),
            name="0-InitPipeline",
        )
        init_input.inputs.recon_all_args = self.parameters["recon_all_args"]
        init_input.inputs.output_dir = str(self.base_dir / self.name / "ReconAll")

        # Run recon-all command
        # FreeSurfer segmentation will be in <subjects_dir>/<image_id>/
        recon_all = npe.Node(interface=ReconAll(), name="1-SegmentationReconAll")
        recon_all.inputs.directive = "all"

        # Generate TSV files containing a summary of the regional statistics
        # in <subjects_dir>/regional_measures
        create_tsv = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "image_id"],
                output_names=["image_id"],
                function=write_tsv_files_task,
            ),
            name="2-CreateTsvFiles",
        )
        connections = [
            (self.input_node, init_input, [("t1w", "t1w")]),
            (
                init_input,
                recon_all,
                [
                    ("subjects_dir", "subjects_dir"),
                    ("t1w", "T1_files"),
                    ("image_id", "subject_id"),
                    ("flags", "flags"),
                ],
            ),
            (init_input, create_tsv, [("subjects_dir", "subjects_dir")]),
            (recon_all, create_tsv, [("subject_id", "image_id")]),
            (create_tsv, self.output_node, [("image_id", "image_id")]),
        ]
        self.connect(connections)



from typing import List, Optional

from pydra import Workflow


def build_eddy_fsl_workflow(
    compute_mask: bool = True,
    image_id: bool = False,
    field: bool = False,
    output_dir: Optional[str] = None,
    name: str = "eddy_fsl",
) -> Workflow:
    """Build a FSL-based pipeline for head motion correction and eddy current distortion correction.

    This pipeline needs FSL as it relies on the Eddy executable.

    Parameters
    ----------
    compute_mask : bool, optional
        If True, the eddy_fsl pipeline computes the brain mask
        of the reference B0 volume using FSL BET. And provide the
        result to Eddy.
        If False, a mask should be provided to the eddy_fsl
        pipeline through the "in_mask" connection.
        Default=True.

    image_id : bool, optional
        Boolean to indicate whether the Eddy node should expect
        a value for its 'out_base' parameter. This is used for
        building the names of the output files.
        Default=False.

    field : bool, optional
        Boolean to indicate whether the Eddy node should expect
        a value for its 'field' input.
        Default=False.

    output_dir: str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.
        TODO: Implement once we can actually write things.

    name : str, optional
        Name of the pipeline. Default='eddy_fsl'.

    Returns
    -------
    Workflow :
        The Pydra workflow.
        This workflow has the following inputs:
            - "dwi_filename": The path to the DWI image
            - "b_vectors_filename": The path to the associated B-vectors file
            - "b_values_filename": The path to the associated B-values file
            - "in_mask": The path to the mask image to be provided to Eddy
              This will be used only if compute_mask is False.
            - "image_id": Prefix to be used for output files
            - "field": The path to the field image to be used by Eddy
            - "reference_b0": The path to the reference B0 image
              This will be used to compute the mask only if compute_mask is True.
            - "total_readout_time": The total readout time extracted from JSON metadata
            - "phase_encoding_direction": The phase encoding direction extracted from JSON metadata

        And the following outputs:
            - "out_parameter": Path to the file storing the output parameters
            - "out_corrected": Path to the corrected image
            - "out_rotated_bvecs": Path to the file holding the rotated B-vectors
    """
    from pydra.tasks import fsl

    from ._tasks import generate_acq_file_task, generate_index_file_task

    wf = Workflow(
        name, input_spec=_build_eddy_fsl_input_specs(image_id, field, compute_mask)
    )

    if compute_mask:
        wf.add(
            fsl.BET(
                name="mask_reference_b0",
                fractional_intensity_threshold=0.3,
                save_brain_mask=True,
                with_robust_brain_center_estimation=True,
                input_image=wf.lzin.reference_b0,
            )
        )

    wf.add(
        generate_acq_file_task(
            dwi_filename=wf.lzin.dwi_filename,
            fsl_phase_encoding_direction=wf.lzin.phase_encoding_direction,
            total_readout_time=wf.lzin.total_readout_time,
            image_id=wf.lzin.image_id if image_id else None,
        )
    )

    wf.add(
        generate_index_file_task(
            b_values_filename=wf.lzin.b_values_filename,
            image_id=wf.lzin.image_id if image_id else None,
        )
    )

    wf.add(
        fsl.Eddy(
            name="eddy_fsl",
            replace_outliers=True,
            bvec_file=wf.lzin.b_vectors_filename,
            bval_file=wf.lzin.b_values_filename,
            input_image=wf.lzin.dwi_filename,
            encoding_file=wf.generate_acq_file_task.lzout.out_file,
            index_file=wf.generate_index_file_task.lzout.out_file,
            brain_mask=wf.mask_reference_b0.lzout.brain_mask
            if compute_mask
            else wf.lzin.in_mask,
            output_basename=wf.lzin.image_id if image_id else "eddy",
        )
    )
    if field:
        wf.eddy_fsl.inputs.fieldmap_image = wf.lzin.field

    wf.set_output(
        {
            "corrected_image": wf.eddy_fsl.lzout.corrected_image,
            "parameters_file": wf.eddy_fsl.lzout.parameters_file,
            "rotated_bvec_file": wf.eddy_fsl.lzout.rotated_bvec_file,
        }
    )

    return wf


def _build_eddy_fsl_input_specs(
    image_id: bool,
    field: bool,
    compute_mask: bool,
) -> List[str]:
    """Returns the input specs for eddy_fsl workflow."""
    fields = [
        "dwi_filename",
        "b_vectors_filename",
        "b_values_filename",
        "total_readout_time",
        "phase_encoding_direction",
    ]
    if image_id:
        fields.append("image_id")
    if field:
        fields.append("field")
    fields.append("reference_b0" if compute_mask else "in_mask")

    return fields



from typing import List, Optional, Tuple

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.t1_volume.dartel2mni.pipeline as pydra_t1_vol_dartel2mni
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-t1-volume-dartel2mni"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.option_group.pipeline_specific_options
@cli_param.option.smooth
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.tissues
@cli_param.option.modulate
@cli_param.option.voxel_size
@option.global_option_group
@option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    smooth: List[int] = (8,),
    tissues: List[int] = (1, 2, 3),
    modulate: bool = True,
    voxel_size: Tuple[float, float, float] = (1.5, 1.5, 1.5),
    n_procs: Optional[int] = None,
) -> None:
    """
    Register DARTEL template to MNI space (Pydra engine).

    GROUP_LABEL is an user-defined identifier to target a specific group of subjects.
    For this pipeline, it is associated to the DARTEL template that you had created
    when running the t1-volume pipeline.
    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Volume/
    """
    from clinica.utils.group import check_group_label

    check_group_label(group_label)
    parameters = {
        "group_label": group_label,
        "tissues": tissues,
        "voxel_size": voxel_size,
        "modulate": modulate,
        "smooth": smooth,
    }
    pipeline = pydra_t1_vol_dartel2mni.build_core_workflow(
        name=pipeline_name,
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(pipeline, n_procs=n_procs)


if __name__ == "__main__":
    cli()



"""Utils to convert AIBL dataset in BIDS."""

from .bids import Modality, paths_to_bids
from .clinical import (
    create_participants_tsv_file,
    create_scans_tsv_file,
    create_sessions_tsv_file,
)

__all__ = [
    "create_participants_tsv_file",
    "create_scans_tsv_file",
    "create_sessions_tsv_file",
    "Modality",
    "paths_to_bids",
]



import os

import numpy as np
import pandas as pd


def load_data(images, caps_directory, subjects, sessions, dataset):
    """

    Args:
        images:
        caps_directory:
        subjects_visits_tsv:
        dataset:

    Returns:
        np 2D array

    """

    df = pd.read_csv(os.path.join(caps_directory), sep="\t")

    all_vector = np.array([])

    participant_id = subjects

    session_id = sessions

    for i in range(len(participant_id)):
        df_sub = df[df.participant_id == participant_id[i]]

        df_analysis = df_sub[[col for col in df_sub.columns if images in col]]

        all_vector = np.append(all_vector, df_analysis.values)
    data = np.zeros((len(participant_id), df_analysis.shape[1]))
    data_temp = np.split(all_vector, len(participant_id))

    for i in range(len(participant_id)):
        for j in range(df_analysis.shape[1]):
            data[i][j] = data_temp[i][j]

    return data



from pathlib import Path
from typing import Callable, Union

from clinica.iotools.bids_utils import StudyName

__all__ = [
    "convert",
    "converter_factory",
    "get_converter_name",
]


def convert(
    study: Union[str, StudyName],
    path_to_dataset: Union[str, Path],
    bids_dir: Union[str, Path],
    path_to_clinical: Union[str, Path],
    **kwargs,
):
    converter_factory(study)(path_to_dataset, bids_dir, path_to_clinical, **kwargs)


def get_converter_name(study: Union[str, StudyName]) -> str:
    study = StudyName(study)
    if study == StudyName.ADNI:
        return "Adni2Bids"
    if study == StudyName.AIBL:
        return "Aibl2Bids"
    if study == StudyName.GENFI:
        return "GenfiToBids"
    if study == StudyName.HABS:
        return "HabsToBids"
    if study == StudyName.NIFD:
        return "Nifd2Bids"
    if study == StudyName.OASIS:
        return "Oasis2Bids"
    if study == StudyName.OASIS3:
        return "Oasis3ToBids"
    if study == StudyName.UKB:
        return "UkbToBids"
    if study == StudyName.IXI:
        return "IxiToBids"


def converter_factory(study: Union[str, StudyName]) -> Callable:
    study = StudyName(study)
    if study == StudyName.ADNI:
        from .adni_to_bids import convert
    if study == StudyName.AIBL:
        from .aibl_to_bids import convert
    if study == StudyName.GENFI:
        from .genfi_to_bids import convert
    if study == StudyName.HABS:
        from .habs_to_bids import convert
    if study == StudyName.NIFD:
        from .nifd_to_bids import convert
    if study == StudyName.OASIS:
        from .oasis_to_bids import convert
    if study == StudyName.OASIS3:
        from .oasis3_to_bids import convert
    if study == StudyName.UKB:
        from .ukb_to_bids import convert
    if study == StudyName.IXI:
        from .ixi_to_bids import convert
    return convert



"""This module contains functions to deal with longitudinal datasets.

Currently, Clinica Pipeline class and clinica/utils folder can not handle the case
where we need to manipulate longitudinal IDs.

When a new longitudinal pipeline will be developed into Clinica, refactoring will be needed.
"""
from pathlib import Path
from typing import List, Optional, Tuple

__all__ = [
    "grab_image_ids_from_caps_directory",
    "move_subjects_dir_to_source_dir",
    "save_part_sess_long_ids_to_tsv",
]


def save_part_sess_long_ids_to_tsv(
    participant_ids: list[str],
    session_ids: list[str],
    long_ids: list[str],
    out_folder: Path,
    file_name: Optional[str] = None,
):
    """Save participant, session and longitudinal IDs to TSV file.

    TODO: Find a way to merge with utils/save_participants_sessions.py::read_participant_tsv into one util
    """
    import pandas as pd

    from clinica.utils.stream import cprint

    out_folder.mkdir(exist_ok=True)
    tsv_file = out_folder / (file_name or "participants.tsv")

    try:
        data = pd.DataFrame(
            {
                "participant_id": participant_ids,
                "session_id": session_ids,
                "long_id": long_ids,
            }
        )
        data.to_csv(tsv_file, sep="\t", index=False, encoding="utf-8")
    except Exception as e:
        cprint(f"Impossible to save {tsv_file} with pandas")
        raise e


def grab_image_ids_from_caps_directory(
    caps_dir: Path,
) -> Tuple[List[str], List[str], List[str]]:
    """Parse CAPS directory to extract participants, sessions and longitudinal IDs.

    Note:
        This function is a simplified version of create_subs_sess_list for CAPS folders with longitudinal IDs

    TODO: Find a way to merge with iotools/utils/data_handling::create_subs_sess_list into one util

    Use case:
    - CAPS
        - subjects
            - sub-CLNC01
                - long-M000M018
                - ses-M000
                - ses-M018
                - ses-M036
            - sub-CLNC02
                - ses-M000
                - ses-M018

    part_ids = ["sub-CLNC01",  "sub-CLNC01",  "sub-CLNC01" ]
    sess_ids = ["ses-M000",     "ses-M018",     "ses-M036"    ]
    long_ids = ["long-M000M018", "long-M000M018", "long-M000M018"]
    (sub-CLNC02 does not have longitudinal ID so it does not appear on the result)

    Parameters
    ----------
    caps_dir : str
        Path to the CAPS directory.
    """
    participants_paths = sorted([f for f in (caps_dir / "subjects").glob("*sub-*")])
    if len(participants_paths) == 0:
        raise IOError("Dataset empty or not CAPS compliant.")
    participant_ids, session_ids, long_ids = [], [], []
    for part_path in participants_paths:
        for ses_path in part_path.glob("*ses-*"):
            for long_path in part_path.glob("*long-*"):
                participant_ids.append(part_path.name)
                session_ids.append(ses_path.name)
                long_ids.append(long_path.name)

    return participant_ids, session_ids, long_ids


def move_subjects_dir_to_source_dir(
    subjects_dir: Path,
    source_dir: Path,
    subject_id: str,
    image_id: Optional[str] = None,
) -> str:
    """
    Move content of `subjects_dir`/`subject_id` to `source_dir`.

    This function will move content of `subject_id` if recon-all has run in $(TMP). This happens when only
    one time point is used. Content of $(TMP)/`subject_id` is copied to `source_dir` before the deletion of $(TMP).

    Parameters
    ----------
    subjects_dir : Path
        $(TMP), if segmentation was performed on 1 time point,
        <base_dir>/<Pipeline.Name>/ReconAll/`subject_id` otherwise

    source_dir : Path
        <base_dir>/<Pipeline.Name>/ReconAll folder.

    subject_id : str
        Subject ID (e.g. "sub-CLNC01_ses-M000" or "sub-CLNC01_ses-M000M018")

    image_id : str, optional

    Returns
    -------
    str :
        subject_id for node connection with Nipype.
    """
    import shutil

    from clinica.utils.stream import cprint

    if str(source_dir) not in str(subjects_dir):
        shutil.copytree(
            src=subjects_dir / subject_id,
            dst=source_dir / (image_id or subject_id) / subject_id,
            symlinks=True,
        )
        shutil.rmtree(subjects_dir)
        cprint(
            msg=(
                f"Segmentation of {subject_id.replace('_', ' | ')} "
                f"has moved to working directory and $SUBJECTS_DIR folder ({subjects_dir}) was deleted."
            ),
            lvl="warning",
        )

    return subject_id



from typing import List, Optional, Tuple

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "t1-volume-existing-template"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.option_group.pipeline_specific_options
@cli_param.option.smooth
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.tissue_classes
@cli_param.option.tissue_probability_maps
@cli_param.option.dont_save_warped_unmodulated
@cli_param.option.save_warped_modulated
@cli_param.option.dartel_tissues
@cli_param.option.tissues
@cli_param.option.modulate
@cli_param.option.voxel_size
@click.pass_context
def cli(
    ctx: click.Context,
    bids_directory: str,
    caps_directory: str,
    group_label: str,
    smooth: List[int] = (8,),
    tissue_classes: List[int] = (1, 2, 3),
    tissue_probability_maps: Optional[str] = None,
    dont_save_warped_unmodulated: bool = False,
    save_warped_modulated: bool = False,
    dartel_tissues: List[int] = (1, 2, 3),
    tissues: List[int] = (1, 2, 3),
    modulate: bool = True,
    voxel_size: Tuple[float, float, float] = (1.5, 1.5, 1.5),
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Volume-based processing of T1-weighted MR images using an existing DARTEL template.

       GROUP_LABEL is an user-defined identifier to target a specific group of subjects.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_Volume/
    """
    from clinica.utils.stream import cprint

    from ..t1_volume_dartel2mni import t1_volume_dartel2mni_cli
    from ..t1_volume_parcellation import t1_volume_parcellation_cli
    from ..t1_volume_register_dartel import t1_volume_register_dartel_cli
    from ..t1_volume_tissue_segmentation import t1_volume_tissue_segmentation_cli

    cprint(
        "The t1-volume-existing-template pipeline is divided into 4 parts:\n"
        "\tt1-volume-tissue-segmentation pipeline: "
        "Tissue segmentation, bias correction and spatial normalization to MNI space\n"
        "\tt1-volume-register-dartel pipeline: "
        "Inter-subject registration using an existing DARTEL template\n"
        "\tt1-volume-dartel2mni pipeline: "
        "DARTEL template to MNI\n"
        "\tt1-volume-parcellation pipeline: "
        "Atlas statistics"
    )

    cprint("Part 1/4: Running t1-volume-segmentation pipeline")

    ctx.invoke(
        t1_volume_tissue_segmentation_cli.cli,
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tissue_classes=tissue_classes,
        dartel_tissues=dartel_tissues,
        tissue_probability_maps=tissue_probability_maps,
        dont_save_warped_unmodulated=dont_save_warped_unmodulated,
        save_warped_modulated=save_warped_modulated,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
    )

    cprint("Part 2/4: Running t1-volume-register-dartel pipeline")
    ctx.invoke(
        t1_volume_register_dartel_cli.cli,
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        group_label=group_label,
        tissues=tissues,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
    )

    cprint("Part 3/4: Running t1-volume-dartel2mni pipeline")
    ctx.invoke(
        t1_volume_dartel2mni_cli.cli,
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        group_label=group_label,
        smooth=smooth,
        tissues=tissues,
        modulate=modulate,
        voxel_size=voxel_size,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
    )

    cprint("Part 4/4: Running t1-volume-parcellation pipeline")
    ctx.invoke(
        t1_volume_parcellation_cli.cli,
        caps_directory=caps_directory,
        group_label=group_label,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
    )


if __name__ == "__main__":
    pass



"""This module contains workflows used by multiple DWI preprocessing pipelines."""

from nipype.pipeline.engine import Workflow

__all__ = ["eddy_fsl_pipeline"]


def eddy_fsl_pipeline(
    base_dir: str,
    use_cuda: bool,
    initrand: bool,
    compute_mask: bool = True,
    image_id: bool = False,
    field: bool = False,
    output_dir=None,
    name: str = "eddy_fsl",
) -> Workflow:
    """Build a FSL-based pipeline for head motion correction and eddy current distortion correction.

    This pipeline needs FSL as it relies on the Eddy executable.

    Parameters
    ----------
    base_dir : str
        The base directory.

    use_cuda : bool
        Boolean to indicate whether cuda should be used or not.

    initrand : bool
        ????

    compute_mask : bool, optional
        If True, the eddy_fsl pipeline computes the brain mask
        of the reference B0 volume using FSL BET. And provide the
        result to Eddy.
        If False, a mask should be provided to the eddy_fsl
        pipeline through the "in_mask" connection.
        Default=True.

    image_id : bool, optional
        Boolean to indicate whether the Eddy node should expect
        a value for its 'out_base' parameter. This is used for
        building the names of the output files.
        Default=False.

    field : bool, optional
        Boolean to indicate whether the Eddy node should expect
        a value for its 'field' input.
        Default=False.

    output_dir: str, optional
        Path to output directory.
        If provided, the pipeline will write its output in this folder.
        Default to None.

    name : str, optional
        Name of the pipeline. Default='eddy_fsl'.

    Returns
    -------
    Workflow :
        The Nipype workflow.
        This workflow has the following inputs:
            - "dwi_filename": The path to the DWI image
            - "b_vectors_filename": The path to the associated B-vectors file
            - "b_values_filename": The path to the associated B-values file
            - "in_mask": The path to the mask image to be provided to Eddy
              This will be used only if compute_mask is False.
            - "image_id": Prefix to be used for output files
            - "field": The path to the field image to be used by Eddy
            - "reference_b0": The path to the reference B0 image
              This will be used to compute the mask only if compute_mask is True.
            - "total_readout_time": The total readout time extracted from JSON metadata
            - "phase_encoding_direction": The phase encoding direction extracted from JSON metadata

        And the following outputs:
            - "out_parameter": Path to the file storing the output parameters
            - "out_corrected": Path to the corrected image
            - "out_rotated_bvecs": Path to the file holding the rotated B-vectors
    """
    import nipype.interfaces.io as nio
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.interfaces.fsl import BET
    from nipype.interfaces.fsl.epi import Eddy

    from .tasks import generate_acq_file_task, generate_index_file_task

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "dwi_filename",
                "b_vectors_filename",
                "b_values_filename",
                "in_mask",
                "image_id",
                "field",
                "reference_b0",
                "total_readout_time",
                "phase_encoding_direction",
            ]
        ),
        name="inputnode",
    )

    mask_reference_b0 = pe.Node(
        BET(frac=0.3, mask=True, robust=True), name="mask_reference_b0"
    )

    generate_acq = pe.Node(
        niu.Function(
            input_names=[
                "dwi_filename",
                "fsl_phase_encoding_direction",
                "total_readout_time",
                "image_id",
                "output_dir",
            ],
            output_names=["out_file"],
            function=generate_acq_file_task,
        ),
        name="generate_acq",
    )
    generate_acq.inputs.output_dir = base_dir

    generate_index = pe.Node(
        niu.Function(
            input_names=["b_values_filename", "image_id", "output_dir"],
            output_names=["out_file"],
            function=generate_index_file_task,
        ),
        name="generate_index",
    )
    generate_index.inputs.output_dir = base_dir

    eddy = pe.Node(interface=Eddy(), name="eddy_fsl")
    eddy.inputs.repol = True
    eddy.inputs.use_cuda = use_cuda
    eddy.inputs.initrand = initrand

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=["out_parameter", "out_corrected", "out_rotated_bvecs"]
        ),
        name="outputnode",
    )

    if output_dir:
        write_results = pe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = output_dir
        write_results.inputs.parameterization = False

    wf = pe.Workflow(name=name)

    connections = [
        (inputnode, generate_acq, [("dwi_filename", "dwi_filename")]),
        (inputnode, generate_acq, [("total_readout_time", "total_readout_time")]),
        (
            inputnode,
            generate_acq,
            [("phase_encoding_direction", "fsl_phase_encoding_direction")],
        ),
        (inputnode, generate_index, [("b_values_filename", "b_values_filename")]),
        (inputnode, eddy, [("b_vectors_filename", "in_bvec")]),
        (inputnode, eddy, [("b_values_filename", "in_bval")]),
        (inputnode, eddy, [("dwi_filename", "in_file")]),
        (generate_acq, eddy, [("out_file", "in_acqp")]),
        (generate_index, eddy, [("out_file", "in_index")]),
        (eddy, outputnode, [("out_parameter", "out_parameter")]),
        (eddy, outputnode, [("out_corrected", "out_corrected")]),
        (eddy, outputnode, [("out_rotated_bvecs", "out_rotated_bvecs")]),
    ]

    if image_id:
        connections += [
            (inputnode, generate_acq, [("image_id", "image_id")]),
            (inputnode, generate_index, [("image_id", "image_id")]),
            (inputnode, eddy, [("image_id", "out_base")]),
        ]

    if field:
        connections += [(inputnode, eddy, [("field", "field")])]

    if compute_mask:
        connections += [
            (inputnode, mask_reference_b0, [("reference_b0", "in_file")]),
            (mask_reference_b0, eddy, [("mask_file", "in_mask")]),
        ]
    else:
        connections += [(inputnode, eddy, [("in_mask", "in_mask")])]

    if output_dir:
        connections += [
            (outputnode, write_results, [("out_parameter", "out_parameter")]),
            (outputnode, write_results, [("out_corrected", "out_corrected")]),
            (outputnode, write_results, [("out_rotated_bvecs", "out_rotated_bvecs")]),
        ]

    wf.connect(connections)

    return wf



import pandas as pd

from clinica.utils.stream import cprint

__all__ = ["get_images_pet"]


def get_images_pet(
    subject: str,
    pet_qc_subj: pd.DataFrame,
    subject_pet_meta: pd.DataFrame,
    df_cols: list[str],
    modality: str,
    sequences_preprocessing_step: list[str],
    viscode_field: str = "VISCODE2",
) -> list[pd.DataFrame]:
    """Selection of scans passing QC and at the chosen preprocessing stage is performed.

    Args:
        subject: Subject identifier
        pet_qc_subj: Dataframe containing QC for scans for the subject
        subject_pet_meta: Dataframe containing metadata for scans for the subject
        df_cols: Columns of output dataframe
        modality: Imaging modality
        sequences_preprocessing_step: List of sequence names that correspond to desired preprocessing stage
        viscode_field: Name of the field in the pet_qc_subj dataframe that provides to the visit code

    Returns: Dataframe containing images metadata
    """
    from clinica.iotools.converter_utils import replace_sequence_chars
    from clinica.utils.pet import Tracer

    subj_dfs_list = []
    for visit in list(pet_qc_subj[viscode_field].unique()):
        if pd.isna(visit):
            continue
        pet_qc_visit = pet_qc_subj[pet_qc_subj[viscode_field] == visit]
        if pet_qc_visit.empty:
            continue
        # If there are several scans for a timepoint we start with image acquired last (higher LONIUID)
        pet_qc_visit = pet_qc_visit.sort_values("LONIUID", ascending=False)

        original_pet_meta = pd.DataFrame(columns=subject_pet_meta.columns)
        qc_visit = pet_qc_visit.iloc[0]
        for qc_index in range(len(pet_qc_visit)):
            qc_visit = pet_qc_visit.iloc[qc_index]

            # We are looking for FDG PET metadata of Original images, that passed QC,
            # acquired at the same date as the current scan that passed QC for the current visit,
            # not containing ‘early’ in the sequence name

            original_pet_meta = subject_pet_meta[
                (subject_pet_meta["Orig/Proc"] == "Original")
                & (subject_pet_meta["Image ID"] == int(qc_visit.LONIUID[1:]))
                & (subject_pet_meta["Scan Date"] == qc_visit.EXAMDATE)
                & ~subject_pet_meta.Sequence.str.contains("early", case=False, na=False)
            ]
            # Check if we found a matching image. If yes, we stop looking for it.
            if not original_pet_meta.empty:
                break

        if original_pet_meta.empty:
            cprint(
                f"No {modality} images metadata for subject {subject} and visit {qc_visit[viscode_field]}",
                lvl="info",
            )
            continue

        original_image = original_pet_meta.iloc[0]

        # Co-registered and Averaged image with the same Series ID of the original image
        averaged_pet_meta = subject_pet_meta[
            subject_pet_meta["Sequence"].isin(sequences_preprocessing_step)
            & (subject_pet_meta["Series ID"] == original_image["Series ID"])
        ]

        # If an explicit Co-registered, Averaged image does not exist,
        # the original image is already in that preprocessing stage.

        if averaged_pet_meta.empty:
            sel_image = original_image
            original = True
        else:
            sel_image = averaged_pet_meta.iloc[0]
            original = False

        phase = "ADNI1" if modality == "PIB-PET" else qc_visit.Phase
        visit = sel_image.Visit
        sequence = replace_sequence_chars(sel_image.Sequence)
        date = sel_image["Scan Date"]
        study_id = sel_image["Study ID"]
        series_id = sel_image["Series ID"]
        image_id = sel_image["Image ID"]

        # If it is an amyloid PET we need to find which is the tracer of the scan and add it to the
        if modality == "Amyloid-PET":
            if "av45" in sel_image.Sequence.lower():
                tracer = Tracer.AV45.value
            elif "fbb" in sel_image.Sequence.lower():
                tracer = Tracer.FBB.value
            else:
                cprint(
                    msg=(
                        f"Unknown tracer for Amyloid PET image metadata for subject {subject} "
                        f"for visit {qc_visit[viscode_field]}"
                    ),
                    lvl="warning",
                )
                continue

            scan_data = [
                [
                    phase,
                    subject,
                    qc_visit[viscode_field],
                    str(visit),
                    sequence,
                    date,
                    str(study_id),
                    str(series_id),
                    str(image_id),
                    original,
                    tracer,
                ]
            ]
        else:
            scan_data = [
                [
                    phase,
                    subject,
                    qc_visit[viscode_field],
                    str(visit),
                    sequence,
                    date,
                    str(study_id),
                    str(series_id),
                    str(image_id),
                    original,
                ]
            ]

        row_to_append = pd.DataFrame(scan_data, columns=df_cols)
        subj_dfs_list.append(row_to_append)

    return subj_dfs_list



"""This module contains utilities to handle groups in Clinica.

See CAPS specifications for details about groups.
"""
from os import PathLike
from typing import List


def check_group_label(group_label: str) -> None:
    """Check that `group_label` is compliant with specifications.

    The group label must be a string composed of alphanumeric
    characters (alphabets or numbers). This means that a group
    label cannot contain spaces nor special characters.

    Parameters
    ----------
    group_label : str
        The group label to verify.

    Raises
    ------
    ValueError :
        If the provided group label is not compliant.
    """
    if not group_label.isalnum():
        raise ValueError(
            "Not valid group_label value: it must be composed only by letters "
            f"and/or numbers (given value: {group_label})."
        )


def _check_group_dir(group_directory: PathLike) -> None:
    from pathlib import Path

    group_directory = Path(group_directory)
    if group_directory.name.startswith("group-"):
        check_group_label(group_directory.name.lstrip("group-"))
    else:
        raise ValueError(f"Group directory {group_directory} is not valid.")


def extract_group_ids(caps_directory: PathLike) -> List[str]:
    """Extract a list of group IDs from a CAPS folder.

    The function searches for sub-folders of `caps_directory/groups`.
    According to the CAPS specifications, these sub-folders should be named
    with their group IDs (e.g. ['group-AD', 'group-HC']).


    Parameters
    ----------
    caps_directory : str
        The CAPS folder to search for group IDs.

    Returns
    -------
    list of str :
        The sorted list of group IDs found inside the provided CAPS folder.

    Raises
    ------
    ValueError :
        If `caps_directory/groups` contains folders which do not
        respect the CAPS naming specifications ("group-{groupID}").
    """
    from pathlib import Path

    caps_directory = Path(caps_directory)

    try:
        groups = [
            group for group in (caps_directory / "groups").iterdir() if group.is_dir()
        ]
    except FileNotFoundError:
        return []
    for group in groups:
        _check_group_dir(group)

    return sorted([group.name for group in groups])



from os import PathLike
from typing import Iterable, Optional, Union

__all__ = ["merge_tsv"]


def merge_tsv(
    bids_directory: Union[str, PathLike],
    output_tsv: Union[str, PathLike],
    caps_directory: Optional[Union[str, PathLike]] = None,
    pipelines: Optional[Iterable[str]] = None,
    volume_atlas_selection: Optional[Iterable[str]] = None,
    freesurfer_atlas_selection: Optional[Iterable[str]] = None,
    pvc_restriction: Optional[int] = None,
    pet_tracers_selection: Optional[Iterable[str]] = None,
    group_selection: Optional[Iterable[str]] = None,
    subjects_sessions_tsv: Optional[Union[str, PathLike]] = None,
    ignore_scan_files: bool = False,
    ignore_session_scan_files: bool = False,
):
    """Merge clinical data into a single TSV file.

    Parameters
    ----------
    bids_directory : PathLike
        The path to the BIDS directory.

    output_tsv : PathLike
        The path to the output TSV file.

    caps_directory : PathLike, optional

    pipelines : Iterable of str, optional
        The names of the pipelines to consider.
        By default, all pipelines will be considered.

    volume_atlas_selection : Iterable of str, optional
        The volume atlases to consider.
        By default, all available atlases will be considered.

    freesurfer_atlas_selection : Iterable of str, optional
        The Freesurfer atlases to consider.
        By default, all available atlases will be considered.

    pvc_restriction : int, optional

    pet_tracers_selection : Iterable of str, optional
        The PET tracers to consider.

    group_selection : Iterable of str, optional
        The group labels to consider.
        By default, all available groups will be considered.

    subjects_sessions_tsv : PathLike, optional

    ignore_scan_files : bool, optional

    ignore_session_scan_files : bool, optional
    """
    from clinica.iotools.utils.data_handling import create_merge_file
    from clinica.utils.inputs import check_bids_folder

    check_bids_folder(bids_directory)

    atlas_selection = []
    if volume_atlas_selection is not None:
        atlas_selection += volume_atlas_selection
    if freesurfer_atlas_selection is not None:
        atlas_selection += freesurfer_atlas_selection
    if group_selection == ():
        group_selection = None
    if pet_tracers_selection == ():
        pet_tracers_selection = None

    create_merge_file(
        bids_directory,
        output_tsv,
        caps_dir=caps_directory,
        pipelines=pipelines,
        ignore_scan_files=ignore_scan_files,
        ignore_sessions_files=ignore_session_scan_files,
        atlas_selection=atlas_selection,
        pvc_restriction=pvc_restriction,
        tsv_file=subjects_sessions_tsv,
        group_selection=group_selection,
        tracers_selection=pet_tracers_selection,
    )



from os import PathLike
from pathlib import Path
from typing import Callable, List, Optional, Union

__all__ = [
    "UserProvidedPath",
    "delete_directories",
    "delete_directories_task",
    "extract_crash_files_from_log_file",
    "extract_image_ids",
    "extract_metadata_from_json",
    "extract_subjects_sessions_from_filename",
    "get_filename_no_ext",
    "get_parent",
    "get_subject_id",
    "load_volume",
    "save_participants_sessions",
    "unzip_nii",
    "zip_nii",
]


UserProvidedPath = Union[str, PathLike]


def zip_nii(in_file: str, same_dir: bool = False) -> str:
    """Compress the provided file(s).

    .. note::
        If a list of file paths is provided, then this function
        will be called on each item and a list of paths to the
        resulting zip files will be returned.
        The type hint is not reflecting this fact because of the
        self-contained requirement from Nipype which is preventing
        the usage of types outside of Python's builtins.

    Parameters
    ----------
    in_file: str
        Path to file to be zipped as a string.

    same_dir: bool, optional
        If True, the zip file is written in the same directory as the
        provided file. Otherwise, it is written in the current working
        directory. Default=False.

    Returns
    -------
    out_file: str
        Path to the resulting zip file.

    Notes
    -----
    If the provided file is already zipped (that is, its extension
    is '.gz'), then nothing is done and the path to the provided file
    is returned.

    Raises
    ------
    FileNotFoundError
        If the provided file does not exist.
    """
    from clinica.utils.filemanip import _zip_unzip_nii  # noqa

    return _zip_unzip_nii(in_file, same_dir, compress=True)


def unzip_nii(
    in_file: str,
    same_dir: bool = False,
) -> str:
    """Decompress the provided file(s).

    .. note::
        If a list of file paths is provided, then this function
        will be called on each item and a list of paths to the
        resulting unzipped files will be returned.
        The type hint is not reflecting this fact because of the
        self-contained requirement from Nipype which is preventing
        the usage of types outside of Python's builtins.

    Parameters
    ----------
    in_file: str
        Path to file to be zipped.

    same_dir: bool, optional
        If True, the unzipped file is written in the same directory as the
        provided zip file. Otherwise, it is written in the current working
        directory. Default=False.

    Returns
    -------
    out_file: str
        Path to the resulting unzipped file.

    Notes
    -----
    If the provided file is not zipped (that is, its extension
    is different from '.gz'), then nothing is done and the path
    to the provided file is returned.

    Raises
    ------
    FileNotFoundError
        If the provided file does not exist.
    """
    from clinica.utils.filemanip import _zip_unzip_nii  # noqa

    return _zip_unzip_nii(in_file, same_dir, compress=False)


def _zip_unzip_nii(in_file: str, same_dir: bool, compress: bool):
    import gzip
    import operator
    import shutil
    from os import getcwd
    from os.path import abspath, join
    from pathlib import Path

    from nipype.utils.filemanip import split_filename

    try:
        # Assuming in_file is path-like.
        in_file = Path(in_file)
    except TypeError:
        try:
            # Assuming in_file is a sequence type.
            return [_zip_unzip_nii(f, same_dir, compress) for f in in_file]
        except TypeError:
            # All other cases.
            return None

    op = operator.eq if compress else operator.ne
    if op(in_file.suffix, ".gz"):
        return str(in_file)

    if not in_file.exists():
        raise FileNotFoundError(f"File {in_file} does not exist.")

    orig_dir, base, ext = split_filename(str(in_file))
    new_ext = ext + ".gz" if compress else ext[:-3]
    out_file = abspath(join(orig_dir if same_dir else getcwd(), base + new_ext))

    outer = open if compress else gzip.open
    inner = gzip.open if compress else open
    with outer(in_file, "rb") as f_in:
        with inner(out_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    return out_file


def load_volume(image_path: str):
    """Load a 3D nifti image from its path.

    If the image is 4D with a dummy fourth dimension,
    then "squeeze" the image into a proper 3D image.

    Parameters
    ----------
    image_path : str
        Path to the image to load.

    Returns
    -------
    img : Nifti1Image
        The loaded 3D image.

    Raises
    ------
    ValueError
        If the loaded image isn't 3D.
    """
    import copy

    import nibabel as nib

    img = nib.load(image_path)
    dim = len(img.shape)
    if dim != 3:
        if dim == 4 and img.shape[3] == 1:
            data = img.get_fdata()
            klass = img.__class__
            header = copy.deepcopy(img.header)
            return klass(data[:, :, :, 0], img.affine, header=header)
        raise ValueError(f"The image is not 3D but {dim}D.")
    return img


def save_participants_sessions(
    participant_ids: list[str],
    session_ids: list[str],
    out_folder: str,
    out_file: Optional[str] = "participants.tsv",
) -> None:
    """Save the participants and sessions to TSV.

    Parameters
    ----------
    participant_ids: List[str]
        List of participant IDs.

    session_ids: List[str]
        List of session IDs.

    out_folder: str
        Folder where the resulting TSV file should be written.

    out_file: str, optional
        Name of the TSV file. Default="participants.tsv".

    Raises
    ------
    ValueError
        If provided `participant_ids` and `session_ids` do not have
        the same length.
    """
    from pathlib import Path

    import pandas as pd

    from clinica.utils.stream import cprint

    if len(participant_ids) != len(session_ids):
        raise ValueError(
            "The number of participant IDs is not equal to the number of session IDs."
        )

    out_folder = Path(out_folder)
    out_folder.mkdir(parents=True, exist_ok=True)
    tsv_file = out_folder / out_file

    data = pd.DataFrame(
        {
            "participant_id": participant_ids,
            "session_id": session_ids,
        }
    )
    try:
        data.to_csv(tsv_file, sep="\t", index=False, encoding="utf-8")
    except Exception as e:
        cprint(msg=f"Impossible to save {out_file} with pandas", lvl="error")
        raise e


def _check_bids_or_caps_compliance(filename: str, sep: str):
    import re

    m = re.search(sep.join([r"(sub-[a-zA-Z0-9]+)", r"(ses-[a-zA-Z0-9]+)"]), filename)
    if not m:
        raise ValueError(
            f"Input filename {filename} is not in a BIDS or CAPS compliant format."
            " It does not contain the subject and session information."
        )

    return m


def get_subject_id(bids_or_caps_file: Union[str, Path]) -> str:
    """Extract the subject ID from a BIDS or CAPS file path.

    The subject ID is defined as

        sub-<participant_id>_ses-<session_label>

    In other words, it is the concatenation of the subject and session labels,
    separated by "_".

    Parameters
    ----------
    bids_or_caps_file: str or Path
        The path to a file from a BIDS or CAPS folder.

    Returns
    -------
    subject_id: str
        The subject ID corresponding to the given file.

    Examples
    --------
    >>> get_subject_id("sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz")
    'sub-01_ses-M000'
    >>> get_subject_id("sub-01_ses-M000_trc-18FAV45_pet.nii.gz")
    Traceback (most recent call last):
    ValueError: Input filename sub-01_ses-M000_trc-18FAV45_pet.nii.gz is not in a BIDS or CAPS compliant format. It does not contain the subject and session information.  # noqa

    See also
    --------
    extract_image_ids
    """
    match = _check_bids_or_caps_compliance(str(bids_or_caps_file), sep="/")
    subject_id = match.group(1) + "_" + match.group(2)

    return subject_id


def get_filename_no_ext(filename: Union[str, Path]) -> str:
    """Get the filename without the extension.

    Parameters
    ----------
    filename: str or Path
        The full filename from which to extract the extension out.

    Returns
    -------
    filename_no_ext: str
        The filename with extension removed.

    Examples
    --------
    >>> get_filename_no_ext("foo.nii.gz")
    'foo'
    >>> get_filename_no_ext("sub-01/ses-M000/sub-01_ses-M000.tar.gz")
    'sub-01_ses-M000'
    """
    if not isinstance(filename, Path):
        filename = Path(filename)
    stem = filename.stem
    while "." in stem:
        stem = Path(stem).stem

    return stem


def extract_image_ids(bids_or_caps_files: list[str]) -> list[str]:
    """Extract the image IDs from a list of BIDS or CAPS files.

    .. warning::
        The image ID is the same as the subject ID from function
        `get_subject_id()` but is extracted from the filename rather
        than from the path. (See examples section).

    Parameters
    ----------
    bids_or_caps_files: List[str]
        List of file paths from which to extract the image IDs.

    Returns
    -------
    id_bids_or_caps_files: List[str]
        List of extracted image IDs.

    Examples
    --------
    >>> extract_image_ids(["sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz"])
    ['sub-01_ses-M000']

    Beware, the image IDs is extracted from the filename and not from the folder names
    as it is the case for `get_subject_id()`:

    >>> extract_image_ids(["foo/bar/baz/sub-foo/ses-bar/foooo/sub-01_ses-M000_foo.json"])
    ['sub-01_ses-M000']

    See also
    --------
    get_subject_id
    """
    id_bids_or_caps_files = []
    for f in bids_or_caps_files:
        match = _check_bids_or_caps_compliance(f, sep="_")
        id_bids_or_caps_files.append(match.group())

    return id_bids_or_caps_files


def extract_subjects_sessions_from_filename(
    bids_or_caps_files: list[str],
) -> tuple[list[str], list[str]]:
    """Extract the subject and session labels from a list of BIDS or CAPS files.

    Parameters
    ----------
    bids_or_caps_files: List[str]
        List of file paths for which to extract the subject and session labels.

    Returns
    -------
    subject_ids: List[str]
        List of subject labels in the same order as `bids_or_caps_files`.

    session_ids: List[str]
        List of session labels in the same order as `bids_or_caps_files`.

    Examples
    --------
    >>> extract_subjects_sessions_from_filename([
    ...     "sub-01/ses-M000/pet/sub-01_ses-M000_trc-18FAV45_pet.nii.gz",
    ...     "foo/bar/baz/sub-foo/ses-bar/foooo/sub-01_ses-M000_foo.json", "sub-01_ses-M000.tar.gz",
    ...])
    (['sub-01', 'sub-01', 'sub-01'], ['ses-M000', 'ses-M000', 'ses-M000'])

    See also
    --------
    extract_image_ids
    """
    id_bids_or_caps_files = extract_image_ids(bids_or_caps_files)
    split = [image_id.split("_") for image_id in id_bids_or_caps_files]
    subject_ids = [p_id[0] for p_id in split]
    session_ids = [s_id[1] for s_id in split]
    return subject_ids, session_ids


def extract_crash_files_from_log_file(filename: str) -> list[str]:
    """Extract crash files (*.pklz) from `filename`.

    Parameters
    ----------
    filename: str
        Path to log file for which the crash files should be extracted.

    Returns
    -------
    crash_files: List[str]
        List of crash files.
    """
    import re
    from pathlib import Path

    filename = Path(filename)
    if not filename.is_file():
        raise ValueError(
            f"extract_crash_files_from_log_file: filename parameter is not a file ({filename})"
        )
    crash_files = []
    with open(filename, "r") as log_file:
        for line in log_file:
            if re.match("(.*)crashfile:(.*)", line):
                crash_files.append(line.replace("\t crashfile:", "").replace("\n", ""))

    return crash_files


def extract_metadata_from_json(
    json_file: Union[str, Path],
    list_keys: list[str],
    handle_missing_keys: Optional[Callable] = None,
) -> list[str]:
    """Extract fields from JSON file.

    Parameters
    ----------
    json_file: str or Path
        Path to a json file containing metadata needed for the pipeline.

    list_keys: list of str
        List of fields the users wants to obtain from the json.

    handle_missing_keys: Optional[Callable]
        Function to use to handle the fields of the json that may be missing.

    Returns
    -------
    list of str:
        Contains the values for the requested fields.
    """
    import json

    from clinica.utils.exceptions import ClinicaException

    try:
        with open(json_file, "r") as file:
            data = json.load(file)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Clinica could not open the following JSON file: {json_file}"
        )
    missing_keys = set(list_keys).difference(set(data.keys()))
    if len(missing_keys) > 0 and handle_missing_keys is None:
        raise ClinicaException(
            f"Clinica could not find the following keys in the following JSON file: {missing_keys}."
        )
    if handle_missing_keys:
        patch = handle_missing_keys(data, missing_keys)
        data.update(patch)
    return [data[k] for k in list_keys]


def get_parent(path: str, n: int = 1) -> Path:
    """Get the path to the nth parent.

    Parameters
    ----------
    path: str
        path to a file.
    n: int
        depth we want to go up in the parents.

    Returns
    -------
    Path
        Path to a parent directory.

    Examples
    --------
    >>> get_parent('/path/to/a/file', 2)
    /path/to
    """
    if n <= 0:
        return Path(path)
    return get_parent(Path(path).parent, n - 1)


def _get_folder_size(folder: Union[str, Path]) -> int:
    """Compute the size in bytes recursively of the given folder.

    Parameters
    ----------
    folder : str or Path
        The path to the folder for which to compute size.

    Returns
    -------
    int :
        The size of the folder in bytes.

    Examples
    --------
    >>> _get_folder_size("./test/instantiation/")
    52571
    """
    import os
    from functools import partial

    prepend = partial(os.path.join, folder)
    return sum(
        [
            (os.path.getsize(f) if os.path.isfile(f) else _get_folder_size(f))
            for f in map(prepend, os.listdir(folder))
        ]
    )


def _get_folder_size_human(folder: Union[str, Path]) -> str:
    """Computes the size of the given folder in human-readable form.

    Parameters
    ----------
    folder : str
        Path to the folder for which to compute size.

    Returns
    -------
    str :
        The size in human-readable form.

    Examples
    --------
    >>> _get_folder_size_human("./test/instantiation/")
    '51.3388671875 KB'
    """
    return _humanize_bytes(_get_folder_size(folder))


def _humanize_bytes(size: int) -> str:
    """Convert a number of bytes in a human-readable form.

    Parameters
    ----------
    size : int
        The number of bytes to convert.

    Returns
    -------
    str :
        The number converted in the best unit for easy reading.
    """
    units = ("B", "KB", "MB", "GB", "TB")

    for unit in units[:-1]:
        if size < 1024.0:
            return f"{size} {unit}"
        size /= 1024.0

    return f"{size} {units[-1]}"


def delete_directories(directories: list[Union[str, Path]]) -> None:
    """This function deletes the directories of the given list".

    Parameters
    ----------
    directories : list of str
        Names of the directories we want to delete.
    """
    import shutil

    total_size: int = 0
    for directory in directories:
        total_size += _get_folder_size(str(directory))
        size = _get_folder_size_human(str(directory))
        shutil.rmtree(directory)
        _print_and_warn(f"Folder {directory} deleted. Freeing {size} of disk space...")
    _print_and_warn(f"Was able to remove {_humanize_bytes(total_size)} of data.")


def _print_and_warn(msg: str, lvl: str = "info") -> None:
    """Print the given message with the given level and warns with the same message."""
    import warnings

    from clinica.utils.stream import cprint

    cprint(msg=msg, lvl=lvl)
    warnings.warn(msg)


def delete_directories_task(directories: list) -> None:
    """Task for Nipype."""
    from clinica.utils.filemanip import delete_directories  # noqa

    return delete_directories(directories)



"""This module contains utilities used by the DWIPreprocessingUsingT1 pipeline."""

import functools
import shutil
from os import PathLike
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np

from clinica.pipelines.dwi.utils import DWIDataset

__all__ = [
    "rename_into_caps",
    "change_itk_transform_type",
    "broadcast_matrix_filename_to_match_b_vector_length",
    "ants_warp_image_multi_transform",
    "rotate_b_vectors",
    "init_input_node",
    "print_end_pipeline",
    "prepare_reference_b0",
    "insert_b0_into_dwi",
    "compute_reference_b0",
    "configure_working_directory",
    "register_b0",
    "extract_sub_ses_folder_name",
]


def rename_into_caps(
    dwi_filename: Path,
    dwi_preproc_filename: Path,
    b_values_preproc_filename: Path,
    b_vectors_preproc_filename: Path,
    b0_brain_mask_filename: Path,
) -> Tuple[str, ...]:
    """Rename the outputs of the pipelines into CAPS.

    Parameters
    ----------
    dwi_filename : str
        The path to the input BIDS DWI to extract the <source_file>.

    dwi_preproc_filename : str
        The path to the preprocessed DWI file.

    dwi_preproc_filename : str
        The path to the preprocessed DWI file.

    b_values_preproc_filename : str
        The path to the preprocessed b-values file.

    b_vectors_preproc_filename : str
        The path to the preprocessed b-vectors file.

    b0_brain_mask_filename : str
        The path to the B0 mask file.

    Returns
    -------
    Tuple[str, str, str, str]
        The different outputs in CAPS format.
    """
    from clinica.pipelines.dwi.utils import rename_files

    return rename_files(
        dwi_filename,
        {
            dwi_preproc_filename: "_space-T1w_desc-preproc_dwi.nii.gz",
            b_values_preproc_filename: "_space-T1w_desc-preproc_dwi.bval",
            b_vectors_preproc_filename: "_space-T1w_desc-preproc_dwi.bvec",
            b0_brain_mask_filename: "_space-T1w_brainmask.nii.gz",
        },
    )


def change_itk_transform_type(input_affine_file: Path) -> Path:
    """Change ITK transform type.

    This function takes in the affine.txt produced by the c3d_affine_tool (which converted
    an FSL FLIRT affine.mat into the affine.txt) it then modifies the 'Transform Type' of
    this affine.txt so that it is compatible with the antsApplyTransforms tool and
    produces a new affine file titled 'updated_affine.txt'.

    Parameters
    ----------
    input_affine_file : Path
        The path to the input affine that should be changed.

    Returns
    -------
    update_affine_file : Path
        The path to the updated affine.
    """
    original_content = input_affine_file.read_text()
    updated_affine_file = input_affine_file.with_name("updated_affine.txt")
    updated_affine_file.write_text(
        original_content.replace(
            "Transform: MatrixOffsetTransformBase_double_3_3",
            "Transform: AffineTransform_double_3_3",
        )
    )

    return updated_affine_file


def broadcast_matrix_filename_to_match_b_vector_length(
    matrix_filename: Path, b_vectors_filename: Path
) -> List[Path]:
    """Return a list of the matrix filename repeated as many times as there are B-vectors."""
    b_vectors = np.loadtxt(b_vectors_filename).T

    return _broadcast_filename_into_list(matrix_filename, len(b_vectors))


def _broadcast_filename_into_list(filename: Path, desired_length: int) -> List[Path]:
    """Return a list of provided filename repeated to match the desired length."""
    return [filename] * desired_length


def ants_warp_image_multi_transform(fix_image, moving_image, ants_warp_affine) -> str:
    import os
    from pathlib import Path

    out_warp = Path("warped_epi.nii.gz").resolve()

    cmd = (
        f"WarpImageMultiTransform 3 {moving_image} {out_warp} -R {fix_image} "
        f"{ants_warp_affine[0]} {ants_warp_affine[1]} {ants_warp_affine[2]}"
    )
    os.system(cmd)

    return str(out_warp)


def rotate_b_vectors(
    b_vectors_filename: Path,
    matrix_filenames: List[str],
    output_dir: Optional[Path] = None,
) -> Path:
    """Rotate the B-vectors contained in the input b_vectors_filename file
    according to the provided list of matrices.

    Parameters
    ----------
    b_vectors_filename : str
        Path to the B-vectors file to rotate.

    matrix_filenames : list of str
        List of paths to rotation matrices to apply to the B-vectors.

    output_dir : str, optional
        If specified, the rotated B-vectors will be written in this folder
        rather than in the same folder as the provided B-vectors.

    Returns
    -------
    rotated_b_vectors_filename : str
        Path to the rotated B-vectors.

    Notes
    -----
    The input affine matrix transforms points in the destination image to their corresponding
    coordinates in the original image. Therefore, this matrix should be inverted first, as
    we want to know the target position of :math:`\\vec{r}`.
    """
    from pathlib import Path

    stem = (
        Path(b_vectors_filename.stem).stem
        if b_vectors_filename.suffix == ".gz"
        else b_vectors_filename.stem
    )
    rotated_b_vectors_filename = b_vectors_filename.with_name(f"{stem}_rotated.bvec")
    if output_dir:
        rotated_b_vectors_filename = output_dir / rotated_b_vectors_filename.name
    else:
        rotated_b_vectors_filename = Path(rotated_b_vectors_filename.name).resolve()

    b_vectors = np.loadtxt(b_vectors_filename).T

    if len(b_vectors) != len(matrix_filenames):
        raise RuntimeError(
            f"Number of b-vectors ({len(b_vectors)}) and rotation "
            f"matrices ({len(matrix_filenames)}) should match."
        )
    rotated_b_vectors = []
    for b_vector, matrix_filename in zip(b_vectors, matrix_filenames):
        if np.all(b_vector == 0.0):
            rotated_b_vectors.append(b_vector)
        else:
            inv_rot = np.linalg.inv(np.loadtxt(matrix_filename))[:3, :3]
            new_b_vector = inv_rot.dot(b_vector)
            rotated_b_vectors.append((new_b_vector / np.linalg.norm(new_b_vector)))

    np.savetxt(rotated_b_vectors_filename, np.array(rotated_b_vectors).T, fmt="%0.15f")

    return rotated_b_vectors_filename


def init_input_node(
    t1w_filename: str,
    dwi_filename: str,
    dwi_json_filename: str,
    b_vectors_filename: str,
    b_values_filename: str,
) -> tuple:
    """Initialize the pipeline.

    Parameters
    ----------
    t1w_filename : str
        The path to the T1w image in BIDS format.

    dwi_filename : str
        The path to the diffusion weighted image in BIDS format.

    dwi_json_filename : str
        The path to the DWI JSON file in BIDS format and containing
        'TotalReadoutTime' and 'PhaseEncodingDirection' metadata
        (see BIDS specifications).

    b_vectors_filename : str
        The path to the b-vectors file in BIDS format.

    b_values_filename : str
        The path of the b-values file in BIDS format.

    Returns
    -------
    image_id : str
        The subject ID extracted from the t1w image path.

    t1w_filename : str
        The path to the T1w image in BIDS format.

    dwi_filename : str
        The path to the diffusion weighted image in BIDS format.

    b_vectors_filename : str
        The path to the b-vectors file in BIDS format.

    b_values_filename : str
        The path of the b-values file in BIDS format.

    total_readout_time : str
        The total readout time extracted from the dwi JSON file.

    phase_encoding_direction : str
        The phase encoding direction for the dwi image, extracted
        from the dwi JSON file.
    """
    from clinica.pipelines.dwi.preprocessing.utils import (
        check_dwi_volume,
        get_readout_time_and_phase_encoding_direction,
    )
    from clinica.pipelines.dwi.utils import DWIDataset
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(t1w_filename)
    check_dwi_volume(
        DWIDataset(
            dwi=dwi_filename,
            b_values=b_values_filename,
            b_vectors=b_vectors_filename,
        )
    )
    (
        total_readout_time,
        phase_encoding_direction,
    ) = get_readout_time_and_phase_encoding_direction(dwi_json_filename)
    print_begin_image(
        image_id,
        ["TotalReadoutTime", "PhaseEncodingDirection"],
        [str(total_readout_time), phase_encoding_direction],
    )

    return (
        image_id,
        t1w_filename,
        dwi_filename,
        b_vectors_filename,
        b_values_filename,
        total_readout_time,
        phase_encoding_direction,
    )


def print_end_pipeline(image_id, final_file):
    """Display end message for `image_id` when `final_file` is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(image_id)


def prepare_reference_b0(
    dwi_dataset: DWIDataset,
    b_value_threshold: float = 5.0,
    working_directory: Optional[Path] = None,
) -> Tuple[Path, DWIDataset]:
    """Prepare reference b=0 image.

    This function prepares the data for further corrections.
    It co-registers the B0 images and then average them in order to
    obtain only one average B0 image.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        DWI dataset for which to prepare the reference B0.

    b_value_threshold : float, optional
        Threshold for B0 volumes. Volumes in the DWI image for which the
        corresponding b_value is <= b_value_threshold will be considered
        as b0 volumes.
        Default=5.0.

    working_directory : Path, optional
        Path to temporary folder where results are stored.
        Defaults to None.

    Returns
    -------
    reference_b0 : Path
        Path to the average of the B0 images or the only B0 image.

    reference_dataset : DWIDataset
        DWI dataset containing the B0 images merged and inserted at index 0.
    """
    from ..utils import check_dwi_dataset

    dwi_dataset = check_dwi_dataset(dwi_dataset)
    working_directory = configure_working_directory(dwi_dataset.dwi, working_directory)
    small_b_dataset, large_b_dataset = _split_dwi_dataset_with_b_values(
        dwi_dataset,
        b_value_threshold=b_value_threshold,
        working_directory=working_directory,
    )
    reference_b0 = compute_reference_b0(
        small_b_dataset.dwi, dwi_dataset.b_values, b_value_threshold, working_directory
    )
    if large_b_dataset:
        return reference_b0, insert_b0_into_dwi(reference_b0, large_b_dataset)
    return reference_b0, small_b_dataset


def insert_b0_into_dwi(b0_filename: PathLike, dwi_dataset: DWIDataset) -> DWIDataset:
    """Inserts a b0 volume into the DWI dataset as the first volume.

    Also updates the b-values and b-vectors files accordingly.

    Parameters
    ----------
    b0_filename : PathLike
        Path to image file containing one b=0 volume (could be the average of a b0 dataset).

    dwi_dataset : DWIDataset
        DWI dataset in which to insert the b0 volume.

    Returns
    -------
    DWIDataset :
        The diffusion dataset : b0 volume + dwi volumes.
    """
    from ..utils import check_dwi_dataset, check_file

    dwi_dataset = check_dwi_dataset(dwi_dataset)
    b0_filename = check_file(b0_filename)

    return DWIDataset(
        dwi=_insert_b0_into_dwi_image(dwi_dataset.dwi, b0_filename),
        b_values=_insert_b0_into_b_values(dwi_dataset.b_values),
        b_vectors=_insert_b0_into_b_vectors(dwi_dataset.b_vectors),
    )


def _insert_b0_into_dwi_image(dwi_filename: Path, b0_filename: Path) -> Path:
    """Insert the provided B0 volume into the DWI image.

    This insertion is done at index 0 along the 4th / time dimension.
    """
    from clinica.utils.image import merge_nifti_images_in_time_dimension

    from ..utils import add_suffix_to_filename

    output_dwi_filename = merge_nifti_images_in_time_dimension(
        (b0_filename, dwi_filename),
        out_file=add_suffix_to_filename(
            _remove_entity_from_filename(dwi_filename, "large_b"),
            "merged",
        ),
    )

    return output_dwi_filename


def _remove_entity_from_filename(filename: Path, entity: str) -> Path:
    """Removes the provided entity from the given file name.

    Parameters
    ----------
    filename : Path
        The path object on which to operate removal of entity.

    entity : str
        The entity which should be removed.

    Returns
    -------
    Path :
        The new Path object without the entity.
    """
    return Path(filename.parent / filename.name.replace(f"_{entity}", ""))


def _insert_b0_into_b_values(b_values_filename: Path) -> Path:
    """Insert a 0 value at index 0 into the b-values file."""
    from ..utils import add_suffix_to_filename

    b_values = np.loadtxt(b_values_filename)
    b_values = np.insert(b_values, 0, 0)
    output_b_values_filename = add_suffix_to_filename(
        _remove_entity_from_filename(b_values_filename, "large_b"), "merged"
    )
    _write_b_values(output_b_values_filename, b_values)

    return output_b_values_filename


def _write_numpy(filename: Path, data: np.ndarray, fmt: str, delimiter: str) -> None:
    """Writes the provided array to the provided filename using the provided formatting."""
    np.savetxt(filename, data, fmt=fmt, delimiter=delimiter)


_write_b_vectors = functools.partial(_write_numpy, fmt="%10.5f", delimiter=" ")
_write_b_values = functools.partial(_write_numpy, fmt="%d", delimiter=" ")


def _insert_b0_into_b_vectors(b_vectors_filename: Path) -> Path:
    """Insert a 0 vector into the b-vectors file."""
    from ..utils import add_suffix_to_filename

    b_vectors = np.loadtxt(b_vectors_filename)
    b_vectors = np.insert(b_vectors, 0, 0.0, axis=1)
    output_b_vectors_filename = add_suffix_to_filename(
        _remove_entity_from_filename(b_vectors_filename, "large_b"), "merged"
    )
    _write_b_vectors(output_b_vectors_filename, b_vectors)

    return output_b_vectors_filename


def compute_reference_b0(
    extracted_b0: Path,
    b_values_filename: Path,
    b_value_threshold: float,
    working_directory: Path,
    clean_working_dir: bool = False,
) -> Path:
    """Compute the reference B0.

    This function calls the b0_flirt FSL pipeline under the hood.
    The pipeline will write files to the provided working directory.

    .. warning::
        If the option clean_working_dir is set to True, this directory
        will be cleaned after execution of the pipeline.

    Parameters
    ----------
    extracted_b0 : Path
        The path to the image of the extracted B0 volume.

    b_values_filename : Path
        The path to the b-values file.

    b_value_threshold : float
        Threshold on b-values to decide whether a DWI volume is
        B0 or not.

    working_directory : Path
        Path to the directory where output files should be written.

    clean_working_dir : bool, optional
        If True, the working directory will be cleaned at the end of
        the execution.
        Default=False.

    Returns
    -------
    registered_b0_filename : Path
        Path to the output image holding the registered B0.

    Raises
    ------
    ValueError :
        If the number of B0 volumes is <= 0.
    """
    from ..utils import compute_average_b0

    if (
        nb_b0s := _count_b0s(
            b_value_filename=b_values_filename, b_value_threshold=b_value_threshold
        )
    ) <= 0:
        raise ValueError(
            f"The number of b0s should be strictly positive (b-val file: {b_values_filename})."
        )
    if nb_b0s == 1:
        return extracted_b0
    registered_b0s = register_b0(
        nb_b0s,
        extracted_b0,
        working_directory=working_directory,
    )
    out_reference_b0 = compute_average_b0(registered_b0s, squeeze=False)
    registered_b0_file_name = extracted_b0.with_name("reference_b0_volume.nii.gz")
    shutil.copy(out_reference_b0, registered_b0_file_name)
    if clean_working_dir:
        shutil.rmtree(working_directory)

    return registered_b0_file_name


def configure_working_directory(
    dwi_filename: Path,
    working_directory: Optional[Path] = None,
) -> Path:
    """Configures a temporary working directory for writing the output files of
    the b0 co-registration.

    Parameters
    ----------
    dwi_filename : Path
        The path to the DWI image file. This is used to create the name of the folder.

    working_directory : Path, optional
        The folder to be used if provided by the user.
        If None, then create a temporary folder to work in.

    Returns
    -------
    Path:
        The path to the configured working directory.
    """
    import hashlib
    import tempfile
    from pathlib import Path

    working_directory = working_directory or Path(tempfile.mkdtemp())
    working_directory = (
        working_directory / hashlib.md5(str(dwi_filename).encode()).hexdigest()
    )
    working_directory.mkdir(parents=True, exist_ok=True)

    return working_directory


def register_b0(
    nb_b0s: int,
    extracted_b0_filename: Path,
    working_directory: Path,
) -> Path:
    """Run the FSL pipeline 'b0_flirt_pipeline' in order to co-register the b0 images.

    This function is a simple wrapper around the b0_flirt_pipeline which configures it,
    runs it, and returns the path to the merged file of interest.

    Parameters
    ----------
    nb_b0s : int
        The number of B0 volumes in the dataset.

    extracted_b0_filename : Path
        The extracted B0 volumes to co-register.

    working_directory : Path
        The working directory in which the pipeline will write intermediary files.

    Returns
    -------
    Path:
        The path to the nifti image containing the co-registered B0 volumes.
        If the pipeline ran successfully, this file should be located in :
        working_directory / b0_coregistration / concat_ref_moving / merged_files.nii.gz
    """
    from .workflows import b0_flirt_pipeline

    b0_flirt = b0_flirt_pipeline(num_b0s=nb_b0s)
    b0_flirt.inputs.inputnode.in_file = str(extracted_b0_filename)
    b0_flirt.base_dir = str(working_directory)
    b0_flirt.run()

    return (
        working_directory
        / "b0_coregistration"
        / "concat_ref_moving"
        / "merged_files.nii.gz"
    )


def extract_sub_ses_folder_name(file_path: str) -> str:
    """This function extracts the name of the folder corresponding to a subject and a session.

    Parameters
    ----------
    file_path: str
        Path to a temporary file for the subject and session of interest.

    Returns
    -------
    str:
    Name of the folder corresponding to a subject and a session.

    Examples
    --------
    >>> extract_sub_ses_folder_name("/localdrive10TB/users/matthieu.joulot/wd/dwi-preprocessing-using-t1/epi_pipeline/4336d63c8556bb56d4e9d1abc617fb3eaa3c38ea/MergeDWIs/Jacobian_image_maths_thresh_merged.nii.gz")
    4336d63c8556bb56d4e9d1abc617fb3eaa3c38ea
    """
    from pathlib import Path

    return (Path(Path(file_path).parent).parent).name


def _split_dwi_dataset_with_b_values(
    dwi_dataset: DWIDataset,
    b_value_threshold: float = 5.0,
    working_directory: Optional[Path] = None,
) -> Tuple[DWIDataset, Optional[DWIDataset]]:
    """Splits the DWI dataset in two through the B-values.

    Splits the DWI volumes into two datasets :
     - the first dataset is relative to volumes having a b-value <= b_value_threshold.
     - the second dataset is relative to volumes having a b-value > b_value_threshold.

    The function writes 6 files (3 files for each dataset), and returns a
    length 2 tuple containing the two DWI datasets.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        DWI image dataset to split.

    b_value_threshold : float, optional
        Defines the b0 volumes as all volumes b-value <= b_value_threshold.
        Defaults to 5.0.

    working_directory : Path, optional
        An optional working directory in which to write the small and large
        DWI datasets. If not provided, they will be written in the same
        directory as the provided DWI dataset.

    Returns
    -------
    small_b_dataset : DWIDataset
        DWI dataset for volumes for which b-value <= b_value_threshold.

    large_b_dataset : DWIDataset
        DWI dataset for volumes for which b-value > b_value_threshold.

    Raises
    ------
    ValueError:
        If b_value_threshold < 0.
    """
    from ..utils import check_b_value_threshold, check_dwi_dataset, get_b0_filter

    check_b_value_threshold(b_value_threshold)
    dwi_dataset = check_dwi_dataset(dwi_dataset)
    b_values, _ = _check_b_values_and_b_vectors(dwi_dataset)
    small_b_filter = get_b0_filter(
        dwi_dataset.b_values, b_value_threshold=b_value_threshold
    )
    large_b_filter = np.array(
        [i for i in range(len(b_values)) if i not in small_b_filter]
    )
    large = None
    if len(small_b_filter) > 0:
        small = _build_dwi_dataset_from_filter(
            dwi_dataset,
            "small_b",
            small_b_filter,
            working_directory=working_directory,
        )
    else:
        raise ValueError("No small dataset found.")
    if len(large_b_filter) > 0:
        large = _build_dwi_dataset_from_filter(
            dwi_dataset,
            "large_b",
            large_b_filter,
            working_directory=working_directory,
        )
    return small, large


def _check_b_values_and_b_vectors(
    dwi_dataset: DWIDataset,
) -> Tuple[np.ndarray, np.ndarray]:
    """Opens the b-values and b-vectors files and transpose the b-vectors if needed."""
    import warnings

    b_values = np.loadtxt(dwi_dataset.b_values, ndmin=1)
    b_vectors = np.loadtxt(dwi_dataset.b_vectors, ndmin=2)
    if b_values.shape[0] == b_vectors.shape[0]:
        warnings.warn(
            "Warning: The b-vectors file should be column-wise. The b-vectors will be transposed",
            UserWarning,
        )
        b_vectors = b_vectors.T

    return b_values, b_vectors


def _build_dwi_dataset_from_filter(
    dwi_dataset: DWIDataset,
    filter_name: str,
    filter_array: np.ndarray,
    working_directory: Optional[Path] = None,
) -> DWIDataset:
    """Builds a new DWI dataset from a given DWI dataset and a filter.

    Parameters
    ----------
    dwi_dataset : DWIDataset
        The DWI dataset to filter.

    filter_name : str
        The name of the filter. This will be used to build the
        file names associated with the new dataset.

    filter_array : np.ndarray
        1D array of indices to filter the DWI dataset.

    working_directory : Path, optional
        An optional working directory in which to write the filtered
        DWI dataset. If not provided, it will be written in the same
        directory as the provided DWI dataset.

    Returns
    -------
    DWIDataset :
        The new filtered DWI dataset.
    """
    from ..utils import check_dwi_dataset

    return check_dwi_dataset(
        DWIDataset(
            dwi=_filter_dwi(
                dwi_dataset,
                filter_name,
                filter_array,
                working_directory=working_directory,
            ),
            b_values=_filter_b_values(
                dwi_dataset,
                filter_name,
                filter_array,
                working_directory=working_directory,
            ),
            b_vectors=_filter_b_vectors(
                dwi_dataset,
                filter_name,
                filter_array,
                working_directory=working_directory,
            ),
        )
    )


def _filter_dwi(
    dwi_dataset: DWIDataset,
    filter_name: str,
    filter_array: np.ndarray,
    working_directory: Optional[Path] = None,
) -> Path:
    """Filters the dwi component of the provided DWI dataset."""
    import nibabel as nib

    from clinica.utils.image import compute_aggregated_volume, get_new_image_like

    from ..utils import add_suffix_to_filename

    dwi_filename = add_suffix_to_filename(dwi_dataset.dwi, filter_name)
    if working_directory:
        dwi_filename = working_directory / dwi_filename.name
    try:
        data = compute_aggregated_volume(
            dwi_dataset.dwi, aggregator=None, volumes_to_keep=filter_array
        )
    except ValueError:
        data = nib.load(dwi_dataset.dwi).get_fdata()
    img = get_new_image_like(dwi_dataset.dwi, data)
    img.to_filename(dwi_filename)

    return dwi_filename


def _filter_b_values(
    dwi_dataset: DWIDataset,
    filter_name: str,
    filter_array: np.ndarray,
    working_directory: Optional[Path] = None,
) -> Path:
    """Filters the b-values component of the provided DWI dataset."""
    from ..utils import add_suffix_to_filename

    b_values, _ = _check_b_values_and_b_vectors(dwi_dataset)
    b_values_filename = add_suffix_to_filename(dwi_dataset.b_values, filter_name)
    if working_directory:
        b_values_filename = working_directory / b_values_filename.name
    _write_b_values(b_values_filename, b_values[filter_array])

    return b_values_filename


def _filter_b_vectors(
    dwi_dataset: DWIDataset,
    filter_name: str,
    filter_array: np.ndarray,
    working_directory: Optional[Path] = None,
) -> Path:
    """Filters the b-vectors component of the provided DWI dataset."""
    from ..utils import add_suffix_to_filename

    _, b_vectors = _check_b_values_and_b_vectors(dwi_dataset)
    b_vectors_filename = add_suffix_to_filename(dwi_dataset.b_vectors, filter_name)
    if working_directory:
        b_vectors_filename = working_directory / b_vectors_filename.name
    _write_b_vectors(b_vectors_filename, np.array([b[filter_array] for b in b_vectors]))

    return b_vectors_filename


def _count_b0s(b_value_filename: PathLike, b_value_threshold: float = 5.0) -> int:
    """Counts the number of volumes where b<=low_bval.

    Parameters
    ----------
    b_value_filename : PathLike
        Path to the b-value file.

    b_value_threshold : float, optional
        Defines the threshold for the b0 volumes as all volumes which
        satisfy b-value <= b_value_threshold.
        Defaults to 5.0.

    Returns
    -------
    num_b0s : int
        Number of b0 volumes.
    """
    from ..utils import get_b0_filter

    return len(get_b0_filter(b_value_filename, b_value_threshold=b_value_threshold))



def perform_suvr_normalization_task(
    pet_image_path: str,
    normalizing_image_path: str,
    reference_mask_path: str,
) -> str:
    from pathlib import Path

    from clinica.pipelines.pet.linear.utils import perform_suvr_normalization

    return str(
        perform_suvr_normalization(
            Path(pet_image_path),
            Path(normalizing_image_path),
            Path(reference_mask_path),
        )
    )


def rename_into_caps_task(
    pet_bids_image_filename: str,
    pet_preprocessed_image_filename: str,
    pet_to_mri_transformation_filename: str,
    suvr_reference_region: str,
    uncropped_image: bool,
    pet_filename_in_t1w_raw: str = None,
    output_dir: str = None,
) -> tuple:
    from pathlib import Path

    from clinica.pipelines.pet.linear.utils import rename_into_caps

    if pet_filename_in_t1w_raw:
        pet_filename_in_t1w_raw = Path(pet_filename_in_t1w_raw)
    if output_dir:
        output_dir = Path(output_dir)
    (
        pet_filename_caps,
        transformation_filename_caps,
        pet_filename_in_t1w_caps,
    ) = rename_into_caps(
        Path(pet_bids_image_filename),
        Path(pet_preprocessed_image_filename),
        Path(pet_to_mri_transformation_filename),
        suvr_reference_region,
        uncropped_image,
        pet_filename_in_t1w_raw,
        output_dir,
    )
    if pet_filename_in_t1w_caps:
        pet_filename_in_t1w_caps = str(pet_filename_in_t1w_caps)
    return (
        str(pet_filename_caps),
        str(transformation_filename_caps),
        pet_filename_in_t1w_caps,
    )



import json
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import IO, List, MutableSequence, NewType, Optional, Union

from attrs import define, fields
from cattr.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattr.preconf.json import make_converter
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from clinica.iotools.bids_dataset_description import get_bids_version
from clinica.utils.bids import BIDS_VERSION
from clinica.utils.exceptions import ClinicaBIDSError, ClinicaCAPSError
from clinica.utils.inputs import DatasetType
from clinica.utils.stream import cprint, log_and_raise, log_and_warn

from .check_dependency import SoftwareDependency

__all__ = [
    "CAPS_VERSION",
    "CAPSDatasetDescription",
    "VersionComparisonPolicy",
    "are_versions_compatible",
    "write_caps_dataset_description",
    "build_caps_dataset_description",
]


CAPS_VERSION = Version("1.0.0")

IsoDate = NewType("IsoDate", datetime)


class VersionComparisonPolicy(str, Enum):
    """Defines the different ways we can compare version numbers in Clinica.

    STRICT: version numbers have to match exactly.
    MINOR : version numbers have to have the same major and minor numbers.
    MAJOR: version numbers only need to share the same major number.
    """

    STRICT = "strict"
    MINOR = "minor"
    MAJOR = "major"


def are_versions_compatible(
    version1: Union[str, Version],
    version2: Union[str, Version],
    policy: Optional[Union[str, VersionComparisonPolicy]] = None,
) -> bool:
    """Returns whether the two provided versions are compatible or not depending on the policy.

    Parameters
    ----------
    version1 : str or Version
        The first version number to compare.

    version2 : str or Version
        The second version number to compare.

    policy : str or VersionComparisonPolicy, optional
        The policy under which to compare version1 with version2.
        By default, a strict policy is used, meaning that version
        numbers have to match exactly.

    Returns
    -------
    bool :
        True if version1 is 'compatible' with version2, False otherwise.
    """
    if isinstance(version1, str):
        version1 = Version(version1)
    if isinstance(version2, str):
        version2 = Version(version2)
    if policy is None:
        policy = VersionComparisonPolicy.STRICT
    else:
        policy = VersionComparisonPolicy(policy)
    if policy == VersionComparisonPolicy.STRICT:
        return version1 == version2
    if policy == VersionComparisonPolicy.MINOR:
        return version1.major == version2.major and version1.minor == version2.minor
    if policy == VersionComparisonPolicy.MAJOR:
        return version1.major == version2.major


@define
class CAPSProcessingDescription:
    """This class models a CAPS processing pipeline metadata.

    Attributes
    ----------
    name : str
        The name of the processing pipeline.
        Example: 't1-linear'.

    date : IsoDate
        The date at which the processing pipeline has been run.
        More precisely, this is the date at which the dataset_description.json
        file is written to disk, which precedes the date at which the pipeline
        finishes processing.

    author : str
        This is the name of the user who ran this processing pipeline.

    machine : str
        This is the name of the machine of which the processing pipeline was run.

    input_path : str
        This is the path to the input dataset.

    dependencies : list of SoftwareDependency
        The dependencies used by the processing.
    """

    name: str
    date: IsoDate
    author: str
    machine: str
    input_path: str
    clinica_version: Version
    dependencies: List[SoftwareDependency]

    @classmethod
    def from_values(
        cls,
        name: str,
        input_path: str,
        dependencies: Optional[List[SoftwareDependency]] = None,
    ):
        from clinica import get_version

        return cls(
            name,
            _get_current_timestamp(),
            _get_username(),
            _get_machine_name(),
            input_path,
            get_version(),
            dependencies or [],
        )

    @classmethod
    def from_dict(cls, values: dict):
        return cls(
            values["Name"],
            values["Date"],
            values["Author"],
            values["Machine"],
            values["InputPath"],
            values["ClinicaVersion"],
            values["Dependencies"],
        )

    def match(
        self,
        processing_name: Optional[str] = None,
        processing_input_path: Optional[str] = None,
    ) -> bool:
        return not (
            (processing_name is not None and self.name != processing_name)
            or (
                processing_input_path is not None
                and (self.input_path != processing_input_path)
            )
        )

    def write(self, to: IO[str]):
        json.dump(converter.unstructure(self), to, indent=4)

    def __str__(self):
        return json.dumps(converter.unstructure(self))

    @classmethod
    def from_file(cls, json_file: Path):
        with open(json_file, "r") as fp:
            content = json.load(fp)
        return converter.structure(content, CAPSProcessingDescription)


@define
class CAPSDatasetDescription:
    """Model representing a CAPS dataset description.

    Attributes
    ----------
    name : str
        The name of the CAPS dataset.

    bids_version : str
        The version number of the BIDS specifications used.

    caps_version : str
        The version number of the CAPS specifications used.

    dataset_type : DatasetType
        The dataset type.

    processing : List of CAPSProcessingDescription
        The list of processing pipelines that have been run.
    """

    name: str
    bids_version: Version = BIDS_VERSION
    caps_version: Version = CAPS_VERSION
    dataset_type: DatasetType = DatasetType.DERIVATIVE
    processing: MutableSequence[CAPSProcessingDescription] = []

    def write(self, to: IO[str]):
        json.dump(converter.unstructure(self), to, indent=4)

    def __str__(self):
        return json.dumps(converter.unstructure(self))

    def has_processing(
        self,
        processing_name: Optional[str] = None,
        processing_input_path: Optional[str] = None,
    ) -> bool:
        return any(
            [
                processing.match(
                    processing_name=processing_name,
                    processing_input_path=processing_input_path,
                )
                for processing in self.processing
            ]
        )

    def get_processing(
        self,
        processing_name: Optional[str] = None,
        processing_input_path: Optional[str] = None,
    ) -> List[CAPSProcessingDescription]:
        return [
            processing
            for processing in self.processing
            if processing.match(
                processing_name=processing_name,
                processing_input_path=processing_input_path,
            )
        ]

    def delete_processing(
        self,
        processing_name: Optional[str] = None,
        processing_input_path: Optional[str] = None,
    ):
        for processing in self.processing:
            if processing.match(
                processing_name=processing_name,
                processing_input_path=processing_input_path,
            ):
                self.processing.remove(processing)

    def add_processing(
        self,
        processing_name: str,
        processing_input_path: str,
        dependencies: Optional[List[SoftwareDependency]] = None,
    ):
        new_processing = CAPSProcessingDescription.from_values(
            processing_name, processing_input_path, dependencies
        )
        existing_processings = self.get_processing(
            processing_name, processing_input_path
        )
        if existing_processings:
            existing_processing = existing_processings[0]
            log_and_warn(
                (
                    f"The CAPS dataset '{self.name}' already has a processing named {processing_name} "
                    f"with an input folder set to {processing_input_path}:\n"
                    f"{existing_processing}\nIt will be overwritten with the following:\n{new_processing}"
                ),
                UserWarning,
            )
            self.delete_processing(existing_processing.name)
        self.processing.append(new_processing)

    @classmethod
    def from_values(
        cls,
        name: Optional[str] = None,
        bids_version: Optional[Version] = None,
        caps_version: Optional[Version] = None,
        processing: Optional[List[CAPSProcessingDescription]] = None,
    ):
        return cls(
            name or _generate_random_name(),
            bids_version or BIDS_VERSION,
            caps_version or CAPS_VERSION,
            DatasetType.DERIVATIVE,
            processing or [],
        )

    @classmethod
    def from_file(cls, json_file: Path):
        with open(json_file, "r") as fp:
            content = json.load(fp)
        return converter.structure(content, CAPSDatasetDescription)

    @classmethod
    def from_dict(cls, values: dict):
        processing = []
        if "Processing" in values:
            processing = [
                CAPSProcessingDescription.from_dict(processing)
                for processing in values["Processing"]
            ]
        return cls(
            values["Name"],
            values["BIDSVersion"],
            values["CAPSVersion"],
            DatasetType(values["DatasetType"]),
            processing,
        )

    def is_compatible_with(
        self, other, policy: Optional[Union[str, VersionComparisonPolicy]] = None
    ) -> bool:
        return are_versions_compatible(
            self.bids_version, other.bids_version, policy=policy
        ) and are_versions_compatible(
            self.caps_version, other.caps_version, policy=policy
        )


def _get_username() -> str:
    import os
    import pwd

    return pwd.getpwuid(os.getuid()).pw_name


def _get_machine_name() -> str:
    import platform

    return platform.node()


def _get_current_timestamp() -> IsoDate:
    return IsoDate(datetime.now())


def _generate_random_name() -> str:
    import uuid

    return str(uuid.uuid4())


def _rename(name: str) -> str:
    """Rename attributes following the specification for the JSON file.

    Basically pascal case with known acronyms such as CAPS fully capitalized.
    """
    return "".join(
        word.upper() if word in ("bids", "caps") else word.capitalize()
        for word in name.split("_")
    )


# Register a JSON converter for the CAPS dataset description model.
converter = make_converter()

# Unstructuring hooks first
converter.register_unstructure_hook(Version, lambda v: str(v))
converter.register_unstructure_hook(SpecifierSet, lambda s: str(s))
converter.register_unstructure_hook(IsoDate, lambda dt: dt.isoformat())


software_dependency_field_renaming = {
    a.name: override(rename=_rename(a.name)) for a in fields(SoftwareDependency)
}
software_dependency_field_renaming_unstructure_hook = make_dict_unstructure_fn(
    SoftwareDependency,
    converter,
    **software_dependency_field_renaming,
)
converter.register_unstructure_hook(
    SoftwareDependency,
    software_dependency_field_renaming_unstructure_hook,
)

caps_processing_field_renaming = {
    a.name: override(rename=_rename(a.name)) for a in fields(CAPSProcessingDescription)
}
caps_processing_field_renaming_unstructure_hook = make_dict_unstructure_fn(
    CAPSProcessingDescription,
    converter,
    **caps_processing_field_renaming,
)
converter.register_unstructure_hook(
    CAPSProcessingDescription,
    caps_processing_field_renaming_unstructure_hook,
)
caps_dataset_description_field_renaming = {
    a.name: override(rename=_rename(a.name)) for a in fields(CAPSDatasetDescription)
}
caps_dataset_field_renaming_unstructure_hook = make_dict_unstructure_fn(
    CAPSDatasetDescription,
    converter,
    **caps_dataset_description_field_renaming,
)
converter.register_unstructure_hook(
    CAPSDatasetDescription,
    caps_dataset_field_renaming_unstructure_hook,
)

# And structuring hooks
converter.register_structure_hook(Version, lambda ts, _: Version(ts))
converter.register_structure_hook(SpecifierSet, lambda s, _: SpecifierSet(s))
converter.register_structure_hook(IsoDate, lambda ts, _: datetime.fromisoformat(ts))

software_dependency_field_renaming_structure_hook = make_dict_structure_fn(
    SoftwareDependency,
    converter,
    **software_dependency_field_renaming,
)
converter.register_structure_hook(
    SoftwareDependency,
    software_dependency_field_renaming_structure_hook,
)

caps_processing_field_renaming_structure_hook = make_dict_structure_fn(
    CAPSProcessingDescription,
    converter,
    **caps_processing_field_renaming,
)
converter.register_structure_hook(
    CAPSProcessingDescription,
    caps_processing_field_renaming_structure_hook,
)
caps_dataset_field_renaming_structure_hook = make_dict_structure_fn(
    CAPSDatasetDescription,
    converter,
    **caps_dataset_description_field_renaming,
)
converter.register_structure_hook(
    CAPSDatasetDescription,
    caps_dataset_field_renaming_structure_hook,
)


def write_caps_dataset_description(
    input_dir: Path,
    output_dir: Path,
    processing_name: str,
    dataset_name: Optional[str] = None,
    bids_version: Optional[str] = None,
    caps_version: Optional[str] = None,
    dependencies: Optional[List[SoftwareDependency]] = None,
) -> None:
    """Write `dataset_description.json` at the root of the CAPS directory.

    Parameters
    ----------
    input_dir : Path
        The path to the folder of the input dataset.
        It can be a BIDS dataset or a CAPS dataset.

    output_dir : Path
        The path to the folder of the output dataset.
        This has to be a CAPS dataset, and this is where
        the requested dataset_description.json file will be written.

    processing_name : str
        The name of the processing performed. By default, pipelines of
        Clinica will set this as the name of the pipeline, but any name
        is possible.

    dataset_name : str, optional
        The name of the CAPS dataset. If not specified, a random identifier will
        be generated. If a dataset_description.json file already exists, the
        existing name will be kept.

    bids_version : str, optional
        The version of the BIDS specifications used.
        By default, this will be set as the BIDS version currently supported by Clinica.

    caps_version : str, optional
        The version of the CAPS specifications used.
        By default, this will be set as the CAPS version currently supported by Clinica.

    dependencies : list of SoftwareDependency
        The dependencies used by the processing.
    """
    description = build_caps_dataset_description(
        input_dir,
        output_dir,
        processing_name,
        dataset_name=dataset_name,
        bids_version=bids_version,
        caps_version=caps_version,
        dependencies=dependencies,
    )
    with open(output_dir / "dataset_description.json", "w") as f:
        description.write(to=f)


def build_caps_dataset_description(
    input_dir: Path,
    output_dir: Path,
    processing_name: str,
    dataset_name: Optional[str] = None,
    bids_version: Optional[str] = None,
    caps_version: Optional[str] = None,
    dependencies: Optional[List[SoftwareDependency]] = None,
) -> CAPSDatasetDescription:
    """Generate the CAPSDatasetDescription for a given CAPS dataset.

    Parameters
    ----------
    input_dir : Path
        The path to the folder of the input dataset.
        It can be a BIDS dataset or a CAPS dataset.

    output_dir : Path
        The path to the folder of the output dataset.
        This has to be a CAPS dataset, and this is where
        the requested dataset_description.json file will be written.

    processing_name : str
        The name of the processing performed. By default, pipelines of
        Clinica will set this as the name of the pipeline, but any name
        is possible.

    dataset_name : str, optional
        The name of the CAPS dataset. If not specified, a random identifier will
        be generated. If a dataset_description.json file already exists, the
        existing name will be kept.

    bids_version : str, optional
        The version of the BIDS specifications used.
        By default, this will be set as the BIDS version currently supported by Clinica.

    caps_version : str, optional
        The version of the CAPS specifications used.
        By default, this will be set as the CAPS version currently supported by Clinica.

    dependencies : list of SoftwareDependency
        The dependencies used by the processing.

    Returns
    -------
    CAPSDatasetDescription :
        The CAPSDatasetDescription generated.
    """
    bids_version_from_input_dir = None
    try:
        bids_version_from_input_dir = get_bids_version(input_dir)
    except ClinicaBIDSError:
        log_and_warn(
            (
                f"Unable to retrieve the BIDS version from the input folder {input_dir}."
                f"Please verify your input dataset. Clinica will assume a BIDS version of {BIDS_VERSION}."
            ),
            UserWarning,
        )
    if (
        bids_version is not None
        and bids_version_from_input_dir is not None
        and bids_version != bids_version_from_input_dir
    ):
        log_and_raise(
            f"The input dataset {input_dir} has BIDS specifications following "
            f"version {bids_version_from_input_dir}, while the BIDS specifications version "
            f"asked for the CAPS creation is {bids_version}. "
            "Please make sure the versions are the same before processing.",
            ClinicaBIDSError,
        )
    new_description = CAPSDatasetDescription.from_values(
        dataset_name, bids_version_from_input_dir, caps_version
    )
    if (output_dir / "dataset_description.json").exists():
        cprint(
            (
                f"The CAPS dataset '{dataset_name}', located at {output_dir}, already "
                "contains a 'dataset_description.json' file."
            ),
            lvl="info",
        )
        previous_description = CAPSDatasetDescription.from_file(
            output_dir / "dataset_description.json"
        )
        if not previous_description.is_compatible_with(
            new_description, VersionComparisonPolicy.STRICT
        ):
            msg = (
                f"Impossible to write the 'dataset_description.json' file in {output_dir} "
                "because it already exists and it contains incompatible metadata."
            )
            log_and_raise(msg, ClinicaCAPSError)
        if previous_description.name != new_description.name:
            log_and_warn(
                (
                    f"The existing CAPS dataset, located at {output_dir} has a "
                    f"name '{previous_description.name}' different from the new "
                    f"name '{new_description.name}'. The old name will be kept."
                ),
                UserWarning,
            )
            new_description.name = previous_description.name
        for processing in previous_description.processing:
            if not processing.match(
                processing_name=processing_name, processing_input_path=str(input_dir)
            ):
                new_description.processing.append(processing)
    new_description.add_processing(processing_name, str(input_dir), dependencies)
    return new_description



from dataclasses import dataclass
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Dict, Tuple, Union

from packaging.version import Version

__all__ = [
    "BIDSLabel",
    "BIDSFileName",
    "BIDS_VERSION",
    "Extension",
    "Suffix",
]

BIDS_VERSION = Version("1.7.0")


class Extension(str, Enum):
    """Possible extensions in BIDS file names."""

    NIIGZ = ".nii.gz"
    NII = ".nii"
    JSON = ".json"
    TSV = ".tsv"
    MAT = ".mat"
    BVAL = ".bval"
    BVEC = ".bvec"


class Suffix(str, Enum):
    """Possible suffixes in BIDS file names."""

    DWI = "dwi"
    PET = "pet"
    T1W = "t1w"
    T2W = "t2w"
    FLAIR = "flair"
    AFFINE = "affine"
    PROBABILITY = "probability"
    DEFORMATION = "deformation"
    PHASEDIFF = "phasediff"
    MAGNITUDE1 = "magnitude1"
    BRAINMASK = "brainmask"
    STATISTICS = "statistics"
    DIFFMODEL = "diffmodel"
    PARCELLATION = "parcellation"


class BIDSLabel(str):
    """A BIDS label is a short string which does not contain symbols
    used for separating entities in BIDS terminology.
    """

    _min_size = 1
    _max_size = 100
    _forbidden_symbols = ("-", "_", ".")

    def __new__(cls, string):
        if not cls._min_size <= len(string) <= cls._max_size:
            raise ValueError(
                f"A string must be between {cls._min_size} and {cls._max_size} to be a valid BIDS label."
            )
        if any([symbol in string for symbol in cls._forbidden_symbols]):
            raise ValueError(
                f"Provided string '{string}' is not a valid BIDS label because "
                f"it contains at least one of these characters: {cls._forbidden_symbols}."
            )
        instance = super().__new__(cls, string)
        return instance


@dataclass
class BIDSFileName:
    """Class modeling a file name following the BIDS specifications."""

    _subject: BIDSLabel
    _session: BIDSLabel
    _suffix: Suffix
    _extension: Extension
    entities: Dict[BIDSLabel, BIDSLabel]

    @property
    def subject(self) -> str:
        return self._subject

    @subject.setter
    def subject(self, subject: str):
        self._subject = BIDSLabel(subject)

    @property
    def session(self) -> str:
        return self._session

    @session.setter
    def session(self, session: str):
        self._session = BIDSLabel(session)

    @property
    def suffix(self) -> str:
        return self._suffix.value

    @suffix.setter
    def suffix(self, suffix: Union[str, Suffix]):
        self._suffix = Suffix(suffix)

    @property
    def extension(self) -> str:
        return self._extension.value

    @extension.setter
    def extension(self, extension: Union[str, Extension]):
        self._extension = Extension(extension)

    @property
    def sub_ses_id(self) -> str:
        return f"sub-{self.subject}_ses-{self.session}"

    @property
    def name(self) -> str:
        if self.entities:
            txt = "_".join([f"{k}-{v}" for k, v in self.entities.items()])
            return f"{self.sub_ses_id}_{txt}_{self.suffix}{self.extension}"
        return f"{self.sub_ses_id}_{self.suffix}{self.extension}"

    @classmethod
    def from_name(cls, filename: Union[str, PathLike]):
        filename, extension = _split_name_from_extension(filename)
        entities, suffix = _tokenize_filename_no_ext(filename)
        subject = entities.pop("sub")
        session = entities.pop("ses")
        return cls(
            BIDSLabel(subject),
            BIDSLabel(session),
            Suffix(suffix),
            Extension(extension),
            {BIDSLabel(k): BIDSLabel(v) for k, v in entities.items()},
        )

    def update_entity(self, entity_name: str, entity_value: str):
        self.entities[BIDSLabel(entity_name)] = BIDSLabel(entity_value)

    def delete_entity(self, entity_name: str):
        entity_name = BIDSLabel(entity_name)
        if entity_name in self.entities:
            self.entities.pop(entity_name)


def _tokenize_filename_no_ext(
    filename_without_extension: str,
) -> Tuple[Dict[str, str], str]:
    if "_" not in filename_without_extension:
        raise ValueError(
            f"BIDS file names have entities separated by '_'. "
            f"You provided {filename_without_extension}."
        )
    tokens = filename_without_extension.split("_")
    if len(tokens) < 3:
        raise ValueError(
            f"A valid BIDS filename should have at least 'sub-XXX_ses-YYY_suffix'. "
            f"You provided {filename_without_extension}."
        )
    suffix = tokens.pop()
    if "-" in suffix:
        raise ValueError(
            f"When tokenizing the filename {filename_without_extension}, the suffix "
            f"found was '{suffix}'. It is invalid because it should not contain a '-' symbol."
        )
    if not all(["-" in token for token in tokens]):
        raise ValueError(
            "The BIDS entities should be key-value pairs separated by a '-' symbol."
            f"The entities found are: {tokens}."
        )
    entities = {k: v for k, v in [s.split("-") for s in tokens]}
    return entities, suffix


def _split_name_from_extension(filename: Union[str, PathLike]) -> Tuple[str, str]:
    extension = ""
    filename = Path(filename)
    while "." in filename.name:
        extension = filename.suffix + extension
        filename = Path(filename.stem)
    return filename.name, extension



import click

from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline
from clinica.pydra import engine_utils

name = "pydra-t1-freesurfer"


@clinica_pipeline
@click.command(name=name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.option_group.pipeline_specific_options
@cli_param.option.atlas_path
def cli(
    bids_directory: str,
    caps_directory: str,
    atlas_path: str,
) -> None:
    """Cross-sectional pre-processing of T1w volumes with FreeSurfer."""
    from . import pipeline

    workflow = pipeline.build_workflow(
        name=name,
        bids_dir=bids_directory,
        caps_dir=caps_directory,
        atlas_path=atlas_path,
    )

    print(engine_utils.run(workflow))


if __name__ == "__main__":
    cli()



"""This module contains the definition of the class DWIPreprocessingPipeline.

All DWI preprocessing pipelines should inherit from this class.
"""

from clinica.pipelines.engine import Pipeline


class DWIPreprocessingPipeline(Pipeline):
    """Class defining common properties of DWI preprocessing pipelines."""

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.stream import cprint

        self.parameters.setdefault("low_bval", 5)
        low_bval = self.parameters["low_bval"]
        if low_bval < 0:
            raise ValueError(
                f"The low_bval is negative ({low_bval}): it should be zero or close to zero."
            )
        if self.parameters["low_bval"] > 100:
            cprint(
                f"The low_bval parameter is {low_bval}: it should be close to zero.",
                lvl="warning",
            )
        self.parameters.setdefault("use_cuda", False)
        self.parameters.setdefault("initrand", False)

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        from clinica.utils.check_dependency import is_binary_present
        from clinica.utils.exceptions import ClinicaMissingDependencyError

        if self.parameters["use_cuda"]:
            if not is_binary_present("eddy_cuda"):
                raise ClinicaMissingDependencyError(
                    "[Error] FSL eddy with CUDA was set but Clinica could "
                    "not find eddy_cuda in your PATH environment. Check that "
                    "https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/UsersGuide#The_eddy_executables "
                    "is correctly set."
                )



import typing as ty
from os import PathLike
from pathlib import Path, PurePath

import pydra
from nipype.interfaces.ants import N4BiasFieldCorrection, RegistrationSynQuick
from pydra import Workflow
from pydra.mark import annotate, task

from clinica.pydra.engine import clinica_io
from clinica.pydra.tasks import download_mni_template_2009c, download_ref_template

n4_bias_field_correction = N4BiasFieldCorrection(bspline_fitting_distance=300)
registration_syn_quick = RegistrationSynQuick(transform_type="a")


@task
@annotate({"return": {"cropped_image": PurePath}})
def crop_image_task(input_image: PathLike, output_dir: PathLike) -> PurePath:
    from pathlib import Path

    from clinica.utils.image import crop_nifti

    return crop_nifti(Path(input_image), Path(output_dir))


@clinica_io
def build_core_workflow(name: str = "core", parameters={}) -> Workflow:
    """Core workflow for the T1 linear pipeline.

    Parameters
    ----------
    name : str
        The name of the workflow.

    Returns
    -------
    Workflow
        The core workflow.
    """
    from pydra.tasks.nipype1.utils import Nipype1Task

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[("T1w", str, {"mandatory": True})],
        bases=(pydra.specs.BaseSpec,),
    )

    wf = Workflow(name, input_spec=input_spec)

    wf.add(download_mni_template_2009c(name="download_mni_template"))

    wf.add(
        Nipype1Task(
            name="n4_bias_field_correction",
            interface=n4_bias_field_correction,
            input_image=wf.lzin.T1w,
        ).split("input_image")
    )

    wf.add(
        Nipype1Task(
            name="registration_syn_quick",
            interface=registration_syn_quick,
            fixed_image=wf.download_mni_template.lzout.mni_template_file,
            moving_image=wf.n4_bias_field_correction.lzout.output_image,
        )
    )

    wf.add(
        crop_image_task(
            name="crop_image",
            interface=crop_image_task,
            input_image=wf.registration_syn_quick.lzout.warped_image,
            output_dir=Path.cwd(),
        )
    )

    wf.set_output(
        [
            ("t1w_corrected_file", wf.n4_bias_field_correction.lzout.output_image),
            ("t1w_registered_file", wf.registration_syn_quick.lzout.warped_image),
            ("t1w_cropped_file", wf.crop_image.lzout.cropped_image),
            ("xfm_file", wf.registration_syn_quick.lzout.out_matrix),
        ]
    )

    return wf



from pathlib import Path
from typing import List, Optional

from nipype import config

from clinica.pipelines.engine import Pipeline

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class T1FreeSurferAtlas(Pipeline):
    """Projection of the results of t1-freesurfer on another atlass

    Returns:
        A clinica pipeline object containing the T1FreeSurfer atlas projection pipeline.
    """

    def __init__(
        self,
        caps_directory: str,
        atlas_path: Optional[str] = None,
    ):
        if atlas_path:
            self.atlas_path = Path(atlas_path)
        else:
            self.atlas_path = None
        super().__init__(caps_directory=caps_directory)

    def _check_custom_dependencies(self) -> None:
        """Check dependencies provided by the developer."""
        pass

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        pass

    @staticmethod
    def get_to_process_with_atlases(
        caps_directory: Path,
        subjects: List[str],
        sessions: List[str],
        atlas_dir_path: Path,
    ) -> List[str]:
        import itertools

        from clinica.pipelines.t1_freesurfer_longitudinal.longitudinal_utils import (
            grab_image_ids_from_caps_directory,
        )
        from clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_correction_utils import (
            get_processed_images,
        )
        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader

        part_ids, sess_ids, list_long_id = grab_image_ids_from_caps_directory(
            caps_directory
        )

        initial_list_to_process = []
        atlas_list = []
        for path in atlas_dir_path.rglob("*rh*6p0.gcs"):
            atlas_name = path.name.split(".")[1].split("_")[0]
            atlas_list.append(atlas_name)

        if caps_directory.is_dir():
            for atlas in atlas_list:
                atlas_info = dict(
                    {
                        "pattern": f"t1/freesurfer_cross_sectional/sub-*_ses-*/stats/rh.{atlas}.stats",
                        "description": f"{atlas}-based segmentation",
                        "needed_pipeline": "t1-freesurfer",
                    }
                )
                t1_freesurfer_longitudinal_output = get_processed_images(
                    caps_directory, part_ids, sess_ids, list_long_id
                )
                t1_freesurfer_longitudinal_output_atlas = get_processed_images(
                    caps_directory, part_ids, sess_ids, list_long_id, atlas
                )
                to_process_long = list(
                    set(t1_freesurfer_longitudinal_output)
                    - set(t1_freesurfer_longitudinal_output_atlas)
                )
                t1_freesurfer_output, _ = clinica_file_reader(
                    subjects, sessions, caps_directory, T1_FS_DESTRIEUX, False
                )
                t1_freesurfer_files, _ = clinica_file_reader(
                    subjects, sessions, caps_directory, atlas_info, False
                )
                image_ids = extract_image_ids(t1_freesurfer_files)
                image_ids_2 = extract_image_ids(t1_freesurfer_output)
                to_process = list(set(image_ids_2) - set(image_ids)) + to_process_long
                initial_list_to_process.append(([atlas], to_process))

        list_to_process = []
        for i in initial_list_to_process:
            if list(itertools.product(i[0], i[1])) != []:
                list_to_process = list_to_process + list(itertools.product(i[0], i[1]))
        return list_to_process

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Notes
        -----
        The list of inputs of the T1FreeSurferAtlas pipeline is:
            * to_process_with_atlases (str): list of the tuples (atlas, sub-ses) to process

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["to_process_with_atlases"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Notes
        -----
        The list of outputs of the T1FreeSurfer pipeline is:
            * image_id (str): Image ID (e.g. sub-CLNC01_ses-M000)

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return ["image_id"]

    def _build_input_node(self):
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        to_process_with_atlases = self.get_to_process_with_atlases(
            self.caps_directory,
            self.subjects,
            self.sessions,
            atlas_dir_path=self.atlas_path,
        )
        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("to_process_with_atlases", to_process_with_atlases),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (
                    read_node,
                    self.input_node,
                    [("to_process_with_atlases", "to_process_with_atlases")],
                )
            ]
        )

    def _build_output_node(self):
        return super().build_output_node()

    def _build_core_nodes(self):
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .tasks import compute_atlas_task, write_tsv_files_task

        # Run an additional two Freesurfer commands if there is an atlas_path specified
        compute_other_atlases = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "caps_directory",
                    "to_process_with_atlases",
                    "path_to_atlas",
                ],
                output_names=["subject_dir", "image_id", "atlas"],
                function=compute_atlas_task,
            ),
            name="1-ComputeOtherAtlases",
        )
        compute_other_atlases.inputs.path_to_atlas = str(self.atlas_path)
        compute_other_atlases.inputs.caps_directory = str(self.caps_directory)

        create_tsv = npe.Node(
            interface=nutil.Function(
                input_names=["subject_dir", "image_id", "atlas"],
                output_names=["image_id"],
                function=write_tsv_files_task,
            ),
            name="2-CreateTsvFiles",
        )

        self.connect(
            [
                (
                    self.input_node,
                    compute_other_atlases,
                    [("to_process_with_atlases", "to_process_with_atlases")],
                ),
                (compute_other_atlases, create_tsv, [("subject_dir", "subject_dir")]),
                (compute_other_atlases, create_tsv, [("image_id", "image_id")]),
                (compute_other_atlases, create_tsv, [("atlas", "atlas")]),
            ]
        )



"""This module contains some functions to be used together with some Nipype functions.

In particular, this module currently contains functions used for Nipype DataSink.
"""


def container_from_filename(bids_or_caps_filename: str) -> str:
    """Extract container from BIDS or CAPS file.

    Parameters
    ----------
    bids_or_caps_filename : str
        Full path to BIDS or CAPS filename.

    Returns
    -------
    str :
        Container path of the form "subjects/<participant_id>/<session_id>".

    Examples
    --------
    >>> from clinica.utils.nipype import container_from_filename
    >>> container_from_filename('/path/to/bids/sub-CLNC01/ses-M000/anat/sub-CLNC01_ses-M000_T1w.nii.gz')
    'subjects/sub-CLNC01/ses-M000'
    >>> container_from_filename('caps/subjects/sub-CLNC01/ses-M000/dwi/preprocessing/sub-CLNC01_ses-M000_preproc.nii')
    'subjects/sub-CLNC01/ses-M000'
    """
    import os
    import re

    m = re.search(r"(sub-[a-zA-Z0-9]+)/(ses-[a-zA-Z0-9]+)", bids_or_caps_filename)
    if not m:
        raise ValueError(
            f"Input filename {bids_or_caps_filename} is not in a BIDS or CAPS compliant format."
            "It does not contain the participant and session ID."
        )
    subject = m.group(1)
    session = m.group(2)
    return os.path.join("subjects", subject, session)


def fix_join(path, *paths):
    """Fix joined path.

    This workaround function is used in pipelines like DWIPreprocessing* or PETVolume. In the workflow.connect part,
    you can use some function that are used as string, causing an import error
    """
    import os

    return os.path.join(path, *paths)
